import os

from cms.utils.compat import DJANGO_3_1

INSTALLED_APPS = [
    "django.contrib.contenttypes",
    "django.contrib.auth",
    "django.contrib.sites",
    "django.contrib.sessions",
    "django.contrib.admin",
    "django.contrib.messages",
    "easy_thumbnails",
    "filer",
    "mptt",
    "cms",
    "menus",
    "treebeard",
    "djangocms_text_ckeditor",
    "djangocms_frontend",
    "djangocms_frontend.contrib.accordion",
    "djangocms_frontend.contrib.alert",
    "djangocms_frontend.contrib.badge",
    "djangocms_frontend.contrib.card",
    "djangocms_frontend.contrib.carousel",
    "djangocms_frontend.contrib.collapse",
    "djangocms_frontend.contrib.content",
    "djangocms_frontend.contrib.grid",
    "djangocms_frontend.contrib.icon",
    "djangocms_frontend.contrib.image",
    "djangocms_frontend.contrib.jumbotron",
    "djangocms_frontend.contrib.link",
    "djangocms_frontend.contrib.listgroup",
    "djangocms_frontend.contrib.media",
    "djangocms_frontend.contrib.navigation",
    "djangocms_frontend.contrib.tabs",
    "djangocms_frontend.contrib.utilities",
    "sekizai",
]

if DJANGO_3_1:
    INSTALLED_APPS += ["django_jsonfield_backport"]

try:  # V4 test?
    import djangocms_versioning  # noqa

    INSTALLED_APPS += [
        "djangocms_versioning",
        "djangocms_alias",
        "djangocms_url_manager",
    ]
except ImportError:  # Nope
    pass

MIDDLEWARE = [
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "cms.middleware.user.CurrentUserMiddleware",
    "cms.middleware.page.CurrentPageMiddleware",
    "cms.middleware.toolbar.ToolbarMiddleware",
    "cms.middleware.language.LanguageCookieMiddleware",
]
CMS_LANGUAGES = {
    1: [
        {
            "code": "en",
            "name": "English",
        }
    ]
}

# required otherwise subject_location would throw an error in the template
THUMBNAIL_PROCESSORS = (
    "easy_thumbnails.processors.colorspace",
    "easy_thumbnails.processors.autocrop",
    "filer.thumbnail_processors.scale_and_crop_with_subject_location",
    "easy_thumbnails.processors.filters",
)

LANGUAGE_CODE = "en"
ALLOWED_HOSTS = ["localhost"]
DJANGOCMS_PICTURE_RESPONSIVE_IMAGES = False
DJANGOCMS_PICTURE_RESPONSIVE_IMAGES_VIEWPORT_BREAKPOINTS = [576, 768, 992]

SECRET_KEY = "fake-key"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [
            os.path.join(os.path.dirname(__file__), "templates"),
            # insert your TEMPLATE_DIRS here
        ],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": "mydatabase",  # This is where you put the name of the db file.
        # If one doesn't exist, it will be created at migration time.
    }
}

CMS_TEMPLATES = (("page.html", "Page"),)

SITE_ID = 1

ROOT_URLCONF = "tests.urls"

DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"

CMS_CONFIRM_VERSION4 = True  # Needed for v4, neglected in v3



from copy import copy

from django import forms
from django.utils.translation import gettext_lazy as _
from entangled.forms import EntangledModelForm

from djangocms_frontend.settings import COLOR_STYLE_CHOICES, DEVICE_SIZES

from ... import settings
from ...common.background import BackgroundFormMixin
from ...common.responsive import ResponsiveFormMixin
from ...common.spacing import MarginFormMixin, PaddingFormMixin
from ...fields import (
    AttributesFormField,
    ButtonGroup,
    ColoredButtonGroup,
    IconGroup,
    TagTypeFormField,
)
from ...helpers import first_choice, link_to_framework_doc
from ...models import FrontendUIItem
from .. import card
from ..grid.constants import GRID_SIZE
from .constants import (
    CARD_ALIGNMENT_CHOICES,
    CARD_INNER_TYPE_CHOICES,
    CARD_LAYOUT_TYPE_CHOICES,
)

# card allow for a transparent color
CARD_COLOR_STYLE_CHOICES = settings.COLOR_STYLE_CHOICES + (
    ("transparent", _("Transparent")),
)

CARD_TEXT_STYLES = COLOR_STYLE_CHOICES + (("white", _("White")),)

mixin_factory = settings.get_forms(card)


class CardLayoutBaseForm(mixin_factory("CardLayout"), EntangledModelForm):
    """
    Components > "Card" Plugin
    https://getbootstrap.com/docs/5.0/components/card/
    """

    class Meta:
        model = FrontendUIItem
        entangled_fields = {
            "config": [
                "card_type",
                "attributes",
            ]
        }
        untangled_fields = (
            "tag_type",
            "create",
        )
        css = settings.ADMIN_CSS

    create = forms.IntegerField(
        label=_("Create cards"),
        help_text=_("Number of cards to create when saving."),
        required=False,
        min_value=0,
        max_value=100,
    )
    card_type = forms.ChoiceField(
        label=_("Card type"),
        choices=CARD_LAYOUT_TYPE_CHOICES,
        initial=first_choice(CARD_LAYOUT_TYPE_CHOICES),
        help_text=link_to_framework_doc("CardLayout", "card_type_link"),
    )
    attributes = AttributesFormField()
    tag_type = TagTypeFormField()


extra_fields_row_cols = {}
for size in settings.DEVICE_SIZES:
    extra_fields_row_cols[f"row_cols_{size}"] = forms.IntegerField(
        label="row-cols" if size == "xs" else f"row-cols-{size}",
        required=False,
        min_value=1,
        max_value=GRID_SIZE,
    )

CardLayoutForm = type(
    "CardLayoutBaseForm",
    (CardLayoutBaseForm,),
    copy(extra_fields_row_cols),
)

CardLayoutForm.Meta.entangled_fields["config"] += extra_fields_row_cols.keys()


class CardForm(
    mixin_factory("Card"),
    BackgroundFormMixin,
    ResponsiveFormMixin,
    MarginFormMixin,
    EntangledModelForm,
):
    """
    Components > "Card" Plugin
    https://getbootstrap.com/docs/5.0/components/card/
    """

    class Meta:
        model = FrontendUIItem
        entangled_fields = {
            "config": [
                "card_alignment",
                "card_outline",
                "card_text_color",
                "card_full_height",
                "attributes",
            ]
        }
        untangled_fields = ("tag_type",)
        css = settings.ADMIN_CSS

    card_outline = forms.ChoiceField(
        label=_("Card outline context"),
        initial=settings.EMPTY_CHOICE[0][0],
        choices=settings.EMPTY_CHOICE + CARD_COLOR_STYLE_CHOICES,
        required=False,
        help_text=_("Uses the border to indicate context."),
        widget=forms.HiddenInput()
        if "card_outline" in getattr(settings, "EXCL_CARD_PROP", ())
        else ColoredButtonGroup(),
    )
    card_alignment = forms.ChoiceField(
        label=_("Alignment"),
        choices=settings.EMPTY_CHOICE + CARD_ALIGNMENT_CHOICES,
        required=False,
        widget=forms.HiddenInput()
        if "card_alignment" in getattr(settings, "EXCL_CARD_PROP", ())
        else IconGroup(),
    )
    card_text_color = forms.ChoiceField(
        label=_("Text context"),
        choices=settings.EMPTY_CHOICE + CARD_TEXT_STYLES,
        required=False,
        widget=forms.HiddenInput()
        if "card_text_color" in getattr(settings, "EXCL_CARD_PROP", ())
        else ColoredButtonGroup(),
    )
    card_full_height = forms.BooleanField(
        label=_("Full height"),
        initial=False,
        required=False,
        help_text=_(
            "If checked cards in one row will automatically extend to the full row height."
        ),
        widget=forms.HiddenInput()
        if "card_full_height" in getattr(settings, "EXCL_CARD_PROP", ())
        else forms.CheckboxInput,
    )
    attributes = AttributesFormField()
    tag_type = TagTypeFormField()


class CardInnerForm(
    mixin_factory("CardInner"),
    BackgroundFormMixin,
    ResponsiveFormMixin,
    PaddingFormMixin,
    EntangledModelForm,
):
    """
    Components > "Card - Inner" Plugin (Header, Footer, Body)
    https://getbootstrap.com/docs/5.0/components/card/
    """

    class Meta:
        model = FrontendUIItem
        entangled_fields = {
            "config": [
                "inner_type",
                "text_alignment",
                "attributes",
            ]
        }
        untangled_fields = ("tag_type",)

    inner_type = forms.ChoiceField(
        label=_("Inner type"),
        choices=CARD_INNER_TYPE_CHOICES,
        initial=first_choice(CARD_INNER_TYPE_CHOICES),
        help_text=_("Define the structure of the plugin."),
        widget=ButtonGroup(attrs=dict(label_class="btn-secondary")),
    )
    text_alignment = forms.ChoiceField(
        label=_("Content alignment"),
        choices=settings.EMPTY_CHOICE + settings.ALIGN_CHOICES,
        required=False,
        widget=forms.HiddenInput()
        if "text_alignment" in getattr(settings, "EXCL_CARD_PROP", ())
        else IconGroup(),
    )
    attributes = AttributesFormField()
    tag_type = TagTypeFormField()


class CardDeckBaseForm(EntangledModelForm):
    class Meta:
        model = FrontendUIItem
        entangled_fields = {
            "config": [
                "attributes",
            ]
        }
        untangled_fields = ("tag_type",)

    attributes = AttributesFormField()
    tag_type = TagTypeFormField()


extra_fields_column = {}
for size in DEVICE_SIZES:
    extra_fields_column[f"{size}_cards"] = forms.IntegerField(
        label="col" if size == "xs" else f"col-{size}",
        required=False,
        min_value=1,
        max_value=GRID_SIZE,
    )

CardDeckForm = type(
    "CardDeckBaseForm",
    (CardDeckBaseForm,),
    copy(extra_fields_column),
)

CardDeckForm.Meta.entangled_fields["config"] += extra_fields_column.keys()



from django import forms
from django.utils.translation import gettext as _
from entangled.forms import EntangledModelForm

from ... import settings
from ...common.spacing import PaddingFormMixin
from ...fields import (
    AttributesFormField,
    ButtonGroup,
    IconGroup,
    TagTypeFormField,
    TemplateChoiceMixin,
)
from ...helpers import first_choice
from ...models import FrontendUIItem
from .constants import (
    TAB_ALIGNMENT_CHOICES,
    TAB_EFFECT_CHOICES,
    TAB_TEMPLATE_CHOICES,
    TAB_TYPE_CHOICES,
)


class TabForm(TemplateChoiceMixin, EntangledModelForm):
    """
    Components > "Navs - Tab" Plugin
    https://getbootstrap.com/docs/5.0/components/navs/
    """

    class Meta:
        model = FrontendUIItem
        entangled_fields = {
            "config": [
                "template",
                "tab_type",
                "tab_alignment",
                "tab_index",
                "tab_effect",
                "attributes",
            ]
        }
        untangled_fields = ("tag_type",)

    template = forms.ChoiceField(
        label=_("Template"),
        choices=TAB_TEMPLATE_CHOICES,
        initial=first_choice(TAB_TEMPLATE_CHOICES),
        help_text=_("This is the template that will be used for the component."),
    )
    tab_type = forms.ChoiceField(
        label=_("Type"),
        choices=TAB_TYPE_CHOICES,
        initial=first_choice(TAB_TYPE_CHOICES),
        widget=ButtonGroup(attrs=dict(property="text")),
    )
    tab_alignment = forms.ChoiceField(
        label=_("Alignment"),
        choices=settings.EMPTY_CHOICE + TAB_ALIGNMENT_CHOICES,
        initial=settings.EMPTY_CHOICE[0][0],
        required=False,
        widget=IconGroup(),
    )
    tab_index = forms.IntegerField(
        label=_("Index"),
        min_value=1,
        required=False,
        help_text=_("Index of element to open on page load starting at 1."),
    )
    tab_effect = forms.ChoiceField(
        label=_("Animation effect"),
        choices=settings.EMPTY_CHOICE + TAB_EFFECT_CHOICES,
        required=False,
    )
    attributes = AttributesFormField()
    tag_type = TagTypeFormField()


class TabItemForm(PaddingFormMixin, EntangledModelForm):
    """
    Components > "Navs - Tab Item" Plugin
    https://getbootstrap.com/docs/5.0/components/navs/
    """

    class Meta:
        model = FrontendUIItem
        entangled_fields = {
            "config": [
                "tab_title",
                "tab_bordered",
                "attributes",
            ]
        }
        untangled_fields = ("tag_type",)

    tab_title = forms.CharField(
        label=_("Tab title"),
    )
    tab_bordered = forms.BooleanField(
        label=_("Bordered"),
        required=False,
        help_text=_("Add borders to the tab item"),
    )

    attributes = AttributesFormField()
    tag_type = TagTypeFormField()



from djangocms_frontend.contrib.carousel.constants import CAROUSEL_DEFAULT_SIZE
from djangocms_frontend.helpers import is_first_child


class CarouselRenderMixin:
    def render(self, context, instance, placeholder):
        instance.add_classes("carousel slide")
        if instance.config.get("carousel_transition", None):
            instance.add_classes(instance.carousel_transition)
        return super().render(context, instance, placeholder)


class CarouselSlideRenderMixin:
    def render(self, context, instance, placeholder):
        parent = instance.parent.get_plugin_instance()[0]
        width = float(context.get("width") or CAROUSEL_DEFAULT_SIZE[0])
        height = float(context.get("height") or CAROUSEL_DEFAULT_SIZE[1])

        if parent.carousel_aspect_ratio:
            aspect_width, aspect_height = tuple(
                int(i) for i in parent.carousel_aspect_ratio.split("x")
            )
            height = width * aspect_height / aspect_width

        instance.add_classes("carousel-item")
        if is_first_child(instance, parent):
            instance.add_classes("active")
        context["link"] = instance.get_link()
        context["aspect_ratio"] = str(width / height)
        context["options"] = {"crop": 10, "size": (width, height), "upscale": True}
        return super().render(context, instance, placeholder)



import warnings

from cms.api import add_plugin
from cms.test_utils.testcases import CMSTestCase

from djangocms_frontend.contrib.grid.cms_plugins import (
    GridColumnPlugin,
    GridContainerPlugin,
    GridRowPlugin,
)
from djangocms_frontend.contrib.grid.forms import (
    GridColumnForm,
    GridContainerForm,
    GridRowForm,
)

from ..fixtures import DJANGO_CMS4, TestFixture


class GridPluginTestCase(TestFixture, CMSTestCase):
    def test_container_plugin(self):
        plugin = add_plugin(
            placeholder=self.placeholder,
            plugin_type=GridContainerPlugin.__name__,
            language=self.language,
        )
        plugin.initialize_from_form(GridContainerForm)
        plugin.save()
        self.assertEqual(
            plugin.plugin_type,
            "GridContainerPlugin",
        )

    def test_grid_row_plugin(self):
        plugin = add_plugin(
            placeholder=self.placeholder,
            plugin_type=GridRowPlugin.__name__,
            language=self.language,
        )
        plugin.initialize_from_form(GridRowForm)
        plugin.save()
        self.assertEqual(
            plugin.plugin_type,
            "GridRowPlugin",
        )

    def test_grid_column_plugin(self):
        plugin = add_plugin(
            placeholder=self.placeholder,
            plugin_type=GridColumnPlugin.__name__,
            language=self.language,
        )
        plugin.initialize_from_form(GridColumnForm)
        plugin.save()
        self.assertEqual(
            plugin.plugin_type,
            "GridColumnPlugin",
        )

    def test_plugin_structure(self):
        container = add_plugin(
            placeholder=self.placeholder,
            plugin_type=GridContainerPlugin.__name__,
            language=self.language,
            config=dict(background_shadow="reg"),
        )
        container.initialize_from_form(GridContainerForm)
        container.save()
        self.publish(self.page, self.language)

        with self.login_user_context(self.superuser):
            response = self.client.get(self.request_url)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "shadow")

        row = add_plugin(
            target=container,
            placeholder=self.placeholder,
            plugin_type=GridRowPlugin.__name__,
            language=self.language,
        )
        row.initialize_from_form(GridRowForm)
        row.save()
        self.publish(self.page, self.language)

        with self.login_user_context(self.superuser):
            response = self.client.get(self.request_url)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, '<div class="row">')

        # add column with values
        plugin = add_plugin(
            target=row,
            placeholder=self.placeholder,
            plugin_type=GridColumnPlugin.__name__,
            language=self.language,
            config=dict(xs_col=12),
        )
        plugin.initialize_from_form(GridColumnForm).save()

        self.publish(self.page, self.language)
        with self.login_user_context(self.superuser):
            response = self.client.get(self.request_url)
        self.assertEqual(response.status_code, 200)
        self.assertTrue(
            ('<div class="col col-12">' in response.content.decode("utf-8"))
            or ('<div class="col-12 col">' in response.content.decode("utf-8")),
            f'<div class="col col-12"> not found in {response.content.decode("utf-8")}',
        )

        # add row without values
        add_plugin(
            target=row,
            placeholder=self.placeholder,
            plugin_type=GridColumnPlugin.__name__,
            language=self.language,
        ).initialize_from_form(GridColumnForm).save()

        self.publish(self.page, self.language)
        with self.login_user_context(self.superuser):
            response = self.client.get(self.request_url)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, '<div class="col">')

    def test_row_plugin_creation(self):
        data = {
            "create": 5,
            "tag_type": "div",
            "margin_devices": ["xs"],
            "padding_devices": ["xl"],
        }

        form = GridRowForm(
            {"config": {"a": 1}, **data}
        )  # GridRowForm & GridColumnForm need config explicitly not empty
        self.assertTrue(form.is_valid())

        if (
            not DJANGO_CMS4
        ):  # TODO: Test for django CMS 4 (failing due to rights issue?)
            request_url = self.get_add_plugin_uri(
                placeholder=self.placeholder,
                plugin_type=GridRowPlugin.__name__,
                language=self.language,
            )
            # create 5 column plugins

            with self.login_user_context(self.superuser), warnings.catch_warnings():
                # hide the "DontUsePageAttributeWarning" warning when using
                # `get_add_plugin_uri` to get cleaner test results
                warnings.simplefilter("ignore")
                response = self.client.post(request_url, data)
            self.assertEqual(response.status_code, 200)
            self.assertContains(response, '<div class="success">')



from cms.api import add_plugin
from cms.test_utils.testcases import CMSTestCase

from djangocms_frontend.contrib.card.cms_plugins import CardInnerPlugin, CardPlugin
from djangocms_frontend.contrib.card.forms import CardForm, CardInnerForm

from ..fixtures import TestFixture


class CardPluginTestCase(TestFixture, CMSTestCase):
    def test_card_plugin(self):
        plugin = add_plugin(
            placeholder=self.placeholder,
            plugin_type=CardPlugin.__name__,
            language=self.language,
        )
        plugin.initialize_from_form(CardForm).save()
        self.publish(self.page, self.language)

        with self.login_user_context(self.superuser):
            response = self.client.get(self.request_url)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, '<div class="card">')

        # add card type
        plugin = add_plugin(
            placeholder=self.placeholder,
            plugin_type=CardPlugin.__name__,
            language=self.language,
            config=dict(
                card_outline="transparent",
                card_alignment="text-start",
                card_text_color="white",
            ),
        )
        plugin.initialize_from_form(CardForm).save()
        self.publish(self.page, self.language)

        with self.login_user_context(self.superuser):
            response = self.client.get(self.request_url)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "border-transparent")
        self.assertContains(response, "text-white")
        self.assertContains(response, "text-start")

        # special case when card outline is given but not card context
        plugin = add_plugin(
            placeholder=self.placeholder,
            plugin_type=CardPlugin.__name__,
            language=self.language,
            config=dict(
                background_context="transparent",
            ),
        )
        plugin.initialize_from_form(CardForm).save()
        self.publish(self.page, self.language)

        with self.login_user_context(self.superuser):
            response = self.client.get(self.request_url)
        self.assertEqual(response.status_code, 200)
        self.assertTrue(
            ('<div class="card bg-transparent">' in response.content.decode("utf-8"))
            or (
                '<div class="bg-transparent card">' in response.content.decode("utf-8")
            ),
            f'<div class="card bg-transparent"> not found in {response.content.decode("utf-8")}',
        )

    def test_card_inner_plugin(self):
        plugin = add_plugin(
            placeholder=self.placeholder,
            plugin_type=CardInnerPlugin.__name__,
            language=self.language,
        )
        plugin.initialize_from_form(CardInnerForm).save()
        self.publish(self.page, self.language)

        with self.login_user_context(self.superuser):
            response = self.client.get(self.request_url)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, '<div class="card-body">')

        # add inner type
        plugin = add_plugin(
            placeholder=self.placeholder,
            plugin_type=CardInnerPlugin.__name__,
            language=self.language,
            config=dict(
                inner_type="card-footer",
            ),
        )
        plugin.initialize_from_form(CardInnerForm).save()
        self.publish(self.page, self.language)

        with self.login_user_context(self.superuser):
            response = self.client.get(self.request_url)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, '<div class="card-footer">')



from cms.api import add_plugin
from cms.test_utils.testcases import CMSTestCase

from djangocms_frontend.contrib.icon.cms_plugins import IconPlugin

from ..fixtures import TestFixture
from .test_models import icon_config


class IconPluginTestCase(TestFixture, CMSTestCase):
    def test_icon_plugin(self):
        add_plugin(
            placeholder=self.placeholder,
            plugin_type=IconPlugin.__name__,
            language=self.language,
            config=icon_config,
            tag_type="i",
        )
        self.publish(self.page, self.language)

        with self.login_user_context(self.superuser):
            response = self.client.get(self.request_url)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, '<i ')
        self.assertContains(response, 'zi-airplane')
        self.assertContains(response, '</i>')
        self.assertContains(response, 'style=')
        self.assertContains(response, '400%')



from django import forms
from django.core.exceptions import ValidationError
from django.utils.translation import gettext as _
from entangled.forms import EntangledModelForm

from ... import settings
from ...common.spacing import SpacingFormMixin
from ...fields import (
    AttributesFormField,
    ButtonGroup,
    ColoredButtonGroup,
    IconGroup,
    TagTypeFormField,
)
from ...helpers import first_choice
from ...models import FrontendUIItem
from .. import utilities

mixin_factory = settings.get_forms(utilities)


class SpacingForm(mixin_factory("Spacing"), EntangledModelForm):
    class Meta:
        model = FrontendUIItem
        entangled_fields = {
            "config": [
                "space_property",
                "space_sides",
                "space_size",
                "space_device",
                "attributes",
            ],
        }
        untangled_fields = ("tag_type",)

    space_property = forms.ChoiceField(
        label=_("Property"),
        choices=settings.SPACER_PROPERTY_CHOICES,
        initial=first_choice(settings.SPACER_PROPERTY_CHOICES),
        widget=ButtonGroup(attrs=dict(property="text")),
    )
    space_sides = forms.ChoiceField(
        label=_("Sides"),
        choices=settings.SPACER_SIDE_CHOICES,
        initial=first_choice(settings.SPACER_SIDE_CHOICES),
        required=False,
        widget=ButtonGroup(attrs=dict(property="text")),
    )
    space_size = forms.ChoiceField(
        label=_("Size"),
        choices=settings.SPACER_SIZE_CHOICES + (("auto", _("Auto")),),
        initial=first_choice(settings.SPACER_SIZE_CHOICES),
        widget=ButtonGroup(attrs=dict(property="text")),
    )
    space_device = forms.ChoiceField(
        label=_("Device"),
        choices=settings.EMPTY_CHOICE + settings.DEVICE_CHOICES,
        initial=settings.EMPTY_CHOICE[0][0],
        required=False,
        widget=IconGroup(),
    )
    attributes = AttributesFormField()
    tag_type = TagTypeFormField()

    def clean(self):
        super().clean()
        if (
            self.cleaned_data["space_property"] == "p"
            and self.cleaned_data["space_size"] == "auto"
        ):
            raise ValidationError(
                {
                    "space_property": _(
                        "Padding does not have an auto spacing. Either "
                        "switch to margin or a defined size."
                    ),
                    "space_size": _(
                        "Padding does not have an auto spacing. Either "
                        "switch to a defined size or change the spacing property."
                    ),
                }
            )


class HeadingForm(mixin_factory("Heading"), SpacingFormMixin, EntangledModelForm):
    class Meta:
        model = FrontendUIItem
        entangled_fields = {
            "config": [
                "heading_level",
                "heading",
                "heading_id",
                "heading_context",
                "heading_alignment",
                "attributes",
            ],
        }
        untangled_fields = ("attributes",)

    HEADINGS = (
        ("h1", _("Heading 1")),
        ("h2", _("Heading 2")),
        ("h3", _("Heading 3")),
        ("h4", _("Heading 4")),
        ("h5", _("Heading 5")),
    )
    ALIGNMENT_CHOICES = (
        ("start", _("Left")),
        ("center", _("Center")),
        ("end", _("Right")),
    )

    heading_level = forms.ChoiceField(
        label=_("Heading level"),
        choices=getattr(settings, "DJANGO_FRONTEND_HEADINGS", HEADINGS),
        required=True,
    )

    heading = forms.CharField(
        label=_("Heading"),
        required=True,
    )

    heading_id = forms.CharField(
        label=_("ID"),
        required=False,
        help_text=_(
            "Fill in unique ID for table of contents. If empty heading will not appear in table of contents."
        ),
    )
    heading_context = forms.ChoiceField(
        label=_("Heading context"),
        required=False,
        choices=settings.EMPTY_CHOICE + settings.COLOR_STYLE_CHOICES,
        initial=settings.EMPTY_CHOICE,
        widget=ColoredButtonGroup(),
    )
    heading_alignment = forms.ChoiceField(
        label=_("Alignment"),
        choices=settings.EMPTY_CHOICE + ALIGNMENT_CHOICES,
        required=False,
        widget=IconGroup(),
    )
    attributes = AttributesFormField()


class TableOfContentsForm(mixin_factory("TableOfContents"), EntangledModelForm):
    class Meta:
        model = FrontendUIItem
        entangled_fields = {
            "config": [
                "list_attributes",
                "link_attributes",
                "attributes",
            ],
        }
        untangled_fields = ()

    list_attributes = AttributesFormField(
        label=_("List attributes"),
        help_text=_(
            "Attributes apply to the <b>list</b> for each level in the table of contents."
        ),
    )

    link_attributes = AttributesFormField(
        label=_("Link attributes"),
        help_text=_(
            "Attributes apply to the <b>link</b> for each entry in the table of contents."
        ),
    )
    attributes = AttributesFormField(
        label=_("Item attributes"),
        help_text=_(
            "Attributes apply to the <b>list items</b> for each entry in the table of contents."
        ),
    )



from cms.api import add_plugin
from cms.test_utils.testcases import CMSTestCase

from djangocms_frontend.contrib.badge.cms_plugins import BadgePlugin
from djangocms_frontend.contrib.badge.forms import BadgeForm

from ..fixtures import TestFixture


class BadgePluginTestCase(TestFixture, CMSTestCase):
    def test_plugin(self):
        plugin = add_plugin(
            placeholder=self.placeholder,
            plugin_type=BadgePlugin.__name__,
            language=self.language,
            config=dict(
                badge_text="some text",
            ),
        )
        plugin.initialize_from_form(BadgeForm).save()
        self.publish(self.page, self.language)

        with self.login_user_context(self.superuser):
            response = self.client.get(self.request_url)
        self.assertEqual(response.status_code, 200)
        self.assertTrue(
            (
                '<span class="badge bg-primary">some text</span>'
                in response.content.decode("utf-8")
            )
            or (
                '<span class="bg-primary badge">some text</span>'
                in response.content.decode("utf-8")
            ),
            f'<span class="badge bg-primary">some text</span> not found in {response.content.decode("utf-8")}',
        )

        # test with pills enabled
        plugin = add_plugin(
            placeholder=self.placeholder,
            plugin_type=BadgePlugin.__name__,
            language=self.language,
            config=dict(
                badge_text="some text",
                badge_pills=True,
            ),
        )
        plugin.initialize_from_form(BadgeForm).save()
        self.publish(self.page, self.language)

        with self.login_user_context(self.superuser):
            response = self.client.get(self.request_url)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "rounded-pill")
        self.assertContains(response, "bg-primary")
        self.assertContains(response, "some text")



from django.utils.translation import gettext_lazy as _

from ...models import FrontendUIItem


class CodeBlock(FrontendUIItem):
    """
    Content > "Code" Plugin
    https://getbootstrap.com/docs/5.0/content/code/
    """

    class Meta:
        proxy = True
        verbose_name = _("Code block")

    def get_short_description(self):
        return f"<{self.code_type}>"


class Blockquote(FrontendUIItem):
    """
    Content > "Blockquote" Plugin
    https://getbootstrap.com/docs/5.0/content/typography/#blockquotes
    """

    class Meta:
        proxy = True
        verbose_name = _("Blockquote")

    def get_short_description(self):
        return self.quote_content


class Figure(FrontendUIItem):
    """
    Content > "Figure" Plugin
    https://getbootstrap.com/docs/5.0/content/figures/
    """

    class Meta:
        proxy = True
        verbose_name = _("Figure")

    def get_short_description(self):
        return self.figure_caption



# original from
# http://tech.octopus.energy/news/2016/01/21/testing-for-missing-migrations-in-django.html
from io import StringIO

from django.core.management import call_command
from django.test import TestCase, override_settings


class MigrationTestCase(TestCase):
    @override_settings(MIGRATION_MODULES={})
    def test_for_missing_migrations(self):
        output = StringIO()
        options = {
            "interactive": False,
            "dry_run": True,
            "stdout": output,
            "check_changes": True,
        }

        try:
            call_command("makemigrations", "djangocms_frontend", **options)
        except SystemExit as e:
            status_code = str(e)
        else:
            # the "no changes" exit code is 0
            status_code = "0"

        if status_code == "1":
            self.fail(f"There are missing migrations:\n {output.getvalue()}")



from django import forms
from django.db.models.fields.related import ManyToOneRel
from django.utils.translation import gettext_lazy as _
from entangled.forms import EntangledModelForm
from filer.fields.image import AdminImageFormField, FilerImageField
from filer.models import Image

from djangocms_frontend.fields import (
    AttributesFormField,
    ButtonGroup,
    TagTypeFormField,
    TemplateChoiceMixin,
)

from ... import settings
from ...common.background import BackgroundFormMixin
from ...fields import HTMLFormField
from ...helpers import first_choice
from ...models import FrontendUIItem
from .. import carousel
from ..link.forms import AbstractLinkForm
from .constants import (
    CAROUSEL_ASPECT_RATIO_CHOICES,
    CAROUSEL_PAUSE_CHOICES,
    CAROUSEL_TEMPLATE_CHOICES,
    CAROUSEL_TRANSITION_CHOICES,
)

mixin_factory = settings.get_forms(carousel)


class CarouselForm(mixin_factory("Carousel"), TemplateChoiceMixin, EntangledModelForm):
    """
    Components > "Carousel" Plugin
    https://getbootstrap.com/docs/5.0/components/carousel/
    """

    class Meta:
        model = FrontendUIItem
        entangled_fields = {
            "config": [
                "template",
                "carousel_aspect_ratio",
                "carousel_controls",
                "carousel_indicators",
                "carousel_interval",
                "carousel_keyboard",
                "carousel_pause",
                "carousel_ride",
                "carousel_wrap",
                "carousel_transition",
                "attributes",
            ]
        }
        untangled_fields = ("tag_type",)

    template = forms.ChoiceField(
        label=_("Template"),
        choices=CAROUSEL_TEMPLATE_CHOICES,
        initial=first_choice(CAROUSEL_TEMPLATE_CHOICES),
        help_text=_("This is the template that will be used for the component."),
    )
    carousel_interval = forms.IntegerField(
        label=_("Interval"),
        initial=5000,
        help_text=_(
            "The amount of time to delay between automatically cycling "
            "an item. If false, carousel will not automatically cycle."
        ),
    )
    carousel_controls = forms.BooleanField(
        label=_("Controls"),
        initial=True,
        required=False,
        help_text=_("Adding in the previous and next controls."),
    )
    carousel_indicators = forms.BooleanField(
        label=_("Indicators"),
        initial=True,
        required=False,
        help_text=_("Adding in the indicators to the carousel."),
    )
    carousel_keyboard = forms.BooleanField(
        label=_("Keyboard"),
        initial=True,
        required=False,
        help_text=_("Whether the carousel should react to keyboard events."),
    )
    carousel_pause = forms.ChoiceField(
        label=_("Pause"),
        choices=CAROUSEL_PAUSE_CHOICES,
        initial=first_choice(CAROUSEL_PAUSE_CHOICES),
        help_text=_(
            'If set to "hover", pauses the cycling of the carousel on '
            '"mouseenter" and resumes the cycling of the carousel on '
            '"mouseleave". If set to "false", hovering over the carousel '
            "won't pause it."
        ),
        widget=ButtonGroup(attrs=dict(property="text")),
    )
    carousel_ride = forms.BooleanField(
        label=_("Auto start"),
        initial=True,
        required=False,
        help_text=_(
            "Autoplays the carousel after the user manually cycles the "
            'first item. If "carousel", autoplays the carousel on load.'
        ),
    )
    carousel_wrap = forms.BooleanField(
        label=_("Wrap"),
        initial=True,
        required=False,
        help_text=_(
            "Whether the carousel should cycle continuously or have " "hard stops."
        ),
    )
    carousel_aspect_ratio = forms.ChoiceField(
        label=_("Aspect ratio"),
        choices=settings.EMPTY_CHOICE + CAROUSEL_ASPECT_RATIO_CHOICES,
        required=False,
        initial=settings.EMPTY_CHOICE[0][0],
        help_text=_(
            "Determines width and height of the image "
            "according to the selected ratio."
        ),
    )
    carousel_transition = forms.ChoiceField(
        label=_("Transition"),
        choices=CAROUSEL_TRANSITION_CHOICES,
        required=False,
        initial=CAROUSEL_TRANSITION_CHOICES[0][0],
        help_text=_("Determines if slides change by sliding or fading."),
        widget=ButtonGroup(attrs=dict(property="text")),
    )
    attributes = AttributesFormField(
        excluded_keys=[
            "id",
            "data-bs-interval",
            "data-bs-keyboard",
            "data-bs-pause",
            "data-bs-ride",
            "data-bs-wrap",
        ],
    )
    tag_type = TagTypeFormField()


class CarouselSlideForm(
    mixin_factory("CarouselSlide"),
    AbstractLinkForm,
    BackgroundFormMixin,
    EntangledModelForm,
):
    """
    Components > "Slide" Plugin
    https://getbootstrap.com/docs/5.0/components/carousel/
    """

    class Meta:
        model = FrontendUIItem
        entangled_fields = {
            "config": [
                "carousel_image",
                "carousel_content",
                "attributes",
            ]
        }
        untangled_fields = ("tag_type",)

    link_is_optional = True

    carousel_image = AdminImageFormField(
        rel=ManyToOneRel(FilerImageField, Image, "id"),
        queryset=Image.objects.all(),
        to_field_name="id",
        label=_("Slide image"),
        required=False,
    )
    carousel_content = HTMLFormField(
        label=_("Content"),
        required=False,
        initial="",
        help_text=_("Content may also be added using child plugins."),
    )
    attributes = AttributesFormField()
    tag_type = TagTypeFormField()



from cms.plugin_pool import plugin_pool
from django.utils.translation import gettext_lazy as _

from ... import settings
from ...cms_plugins import CMSUIPlugin
from ...common.attributes import AttributesMixin
from .. import collapse
from . import forms, models

mixin_factory = settings.get_renderer(collapse)


@plugin_pool.register_plugin
class CollapsePlugin(mixin_factory("Collapse"), AttributesMixin, CMSUIPlugin):
    """
    Component > "Collapse" Plugin
    https://getbootstrap.com/docs/5.0/components/collapse/
    """

    name = _("Collapse")
    module = _("Frontend")
    model = models.Collapse
    form = forms.CollapseForm
    change_form_template = "djangocms_frontend/admin/collapse.html"
    allow_children = True
    child_classes = [
        "CollapseTriggerPlugin",
        "CollapseContainerPlugin",
        "LinkPlugin",
        "CardPlugin",
        "SpacingPlugin",
        "GridRowPlugin",
    ]

    fieldsets = [
        (None, {"fields": ("collapse_siblings",)}),
    ]


@plugin_pool.register_plugin
class CollapseTriggerPlugin(
    mixin_factory("CollapseTrigger"), AttributesMixin, CMSUIPlugin
):
    """
    Component > "Collapse" Plugin
    https://getbootstrap.com/docs/5.0/components/collapse/
    """

    name = _("Collapse trigger")
    module = _("Frontend")
    model = models.CollapseTrigger
    form = forms.CollapseTriggerForm
    allow_children = True
    parent_classes = [
        "CardPlugin",
        "CardInnerPlugin",
        "CollapsePlugin",
        "GridColumnPlugin",
    ]

    fieldsets = [
        (None, {"fields": ("trigger_identifier",)}),
    ]


@plugin_pool.register_plugin
class CollapseContainerPlugin(
    mixin_factory("CollapseContainer"), AttributesMixin, CMSUIPlugin
):
    """
    Component > "Collapse Container" Plugin
    https://getbootstrap.com/docs/5.0/components/collapse/
    """

    name = _("Collapse container")
    module = _("Frontend")
    model = models.CollapseContainer
    form = forms.CollapseContainerForm
    allow_children = True
    parent_classes = [
        "CardPlugin",
        "CardInnerPlugin",
        "CollapsePlugin",
        "GridColumnPlugin",
    ]

    fieldsets = [
        (None, {"fields": ("container_identifier",)}),
    ]



from django import forms
from django.utils.translation import gettext_lazy as _
from entangled.forms import EntangledModelFormMixin

from djangocms_frontend import settings
from djangocms_frontend.fields import ButtonGroup, ColoredButtonGroup
from djangocms_frontend.helpers import first_choice, insert_fields


class BackgroundMixin:
    def get_fieldsets(self, request, obj=None):
        return insert_fields(
            super().get_fieldsets(request, obj),
            (
                "background_context",
                ("background_opacity", "background_shadow"),
            ),
            block=None,
            position=-1,
            blockname=_("Background"),
        )

    def render(self, context, instance, placeholder):
        if getattr(instance, "background_context", ""):
            instance.add_classes(f"bg-{instance.background_context}")
        if getattr(instance, "background_opacity", "100") != "100":
            if instance.background_opacity:
                instance.add_classes(f"bg-opacity-{instance.background_opacity}")
        if getattr(instance, "background_shadow", ""):
            if instance.background_shadow == "reg":
                instance.add_classes("shadow")
            else:
                instance.add_classes(f"shadow-{instance.background_shadow}")
        return super().render(context, instance, placeholder)


class BackgroundFormMixin(EntangledModelFormMixin):
    class Meta:
        entangled_fields = {
            "config": [
                "background_context",
                "background_opacity",
                "background_shadow",
            ]
        }

    background_context = forms.ChoiceField(
        label=_("Background context"),
        required=False,
        choices=settings.EMPTY_CHOICE
        + settings.COLOR_STYLE_CHOICES
        + (("transparent", _("Transparent")),),
        initial=settings.EMPTY_CHOICE[0][0],
        widget=ColoredButtonGroup(),
    )
    background_opacity = forms.ChoiceField(
        label=_("Background opacity"),
        required=False,
        choices=settings.framework_settings.OPACITY_CHOICES,
        initial=first_choice(settings.framework_settings.OPACITY_CHOICES),
        widget=ButtonGroup(attrs=dict(property="opacity")),
        help_text=_("Opacity of card background color (only if no outline selected)"),
    )

    background_shadow = forms.ChoiceField(
        label=_("Shadow"),
        required=False,
        choices=settings.EMPTY_CHOICE + settings.framework_settings.SHADOW_CHOICES,
        initial=settings.EMPTY_CHOICE[0][0],
        widget=ButtonGroup(attrs=dict(property="shadow")),
        help_text=_("Use shadows to optically lift cards from the background."),
    )



from django.test import TestCase

from djangocms_frontend.contrib.card.forms import CardForm, CardInnerForm
from djangocms_frontend.contrib.card.models import Card, CardInner


class CardModelTestCase(TestCase):
    def test_card_instance(self):
        instance = Card.objects.create().initialize_from_form(CardForm)
        self.assertEqual(str(instance), "Card (1)")
        self.assertEqual(instance.get_short_description(), "")
        instance.config["background_context"] = "primary"
        self.assertEqual(instance.get_short_description(), ".bg-primary")
        instance.config["card_outline"] = "secondary"
        self.assertEqual(
            instance.get_short_description(), ".bg-primary .border-secondary"
        )
        instance.card_alignment = "center"
        self.assertEqual(
            instance.get_short_description(), ".bg-primary .border-secondary .center"
        )

    def test_card_inner_instance(self):
        instance = CardInner.objects.create().initialize_from_form(CardInnerForm)
        self.assertEqual(str(instance), "CardInner (1)")
        self.assertEqual(instance.get_short_description(), "(body)")



from cms.plugin_pool import plugin_pool
from django.utils.translation import gettext_lazy as _

from ... import settings
from ...cms_plugins import CMSUIPlugin
from ...common.attributes import AttributesMixin
from ...common.responsive import ResponsiveMixin
from .. import media
from . import forms, models

mixin_factory = settings.get_renderer(media)


@plugin_pool.register_plugin
class MediaPlugin(
    mixin_factory("Media"), AttributesMixin, ResponsiveMixin, CMSUIPlugin
):
    """
    Layout > "Media" Plugin
    http://getbootstrap.com/docs/4.0/layout/media-object/
    """

    name = _("Media")
    module = _("Frontend")
    model = models.Media
    form = forms.MediaForm
    change_form_template = "djangocms_frontend/admin/media.html"
    allow_children = True

    fieldsets = settings.EMPTY_FIELDSET


@plugin_pool.register_plugin
class MediaBodyPlugin(mixin_factory("MediaBody"), AttributesMixin, CMSUIPlugin):
    """
    Layout > "Media body" Plugin
    http://getbootstrap.com/docs/4.0/layout/media-object/
    """

    name = _("Media body")
    module = _("Frontend")
    model = models.MediaBody
    form = forms.MediaBodyForm
    change_form_template = "djangocms_frontend/admin/media.html"
    allow_children = True
    parent_classes = ["MediaPlugin"]

    fieldsets = settings.EMPTY_FIELDSET



from importlib import import_module

from django.apps import AppConfig
from django.conf import settings
from django.urls import NoReverseMatch, clear_url_caches, include, path, reverse
from django.utils.translation import gettext_lazy as _


class LinkConfig(AppConfig):
    name = "djangocms_frontend.contrib.link"
    verbose_name = _("Link")

    def ready(self):
        """Install the URLs"""
        try:
            reverse("dcf_autocomplete:ac_view")
        except NoReverseMatch:  # Not installed yet
            urlconf_module = import_module(settings.ROOT_URLCONF)
            urlconf_module.urlpatterns = [
                path(
                    "@dcf-frontend_link/",
                    include(
                        "djangocms_frontend.contrib.link.urls",
                        namespace="dcf_autocomplete",
                    ),
                )
            ] + urlconf_module.urlpatterns
            clear_url_caches()



from django import forms
from django.utils.translation import gettext_lazy as _
from entangled.forms import EntangledModelForm

from djangocms_frontend.fields import (
    AttributesFormField,
    ColoredButtonGroup,
    TagTypeFormField,
)
from djangocms_frontend.helpers import first_choice
from djangocms_frontend.models import FrontendUIItem
from djangocms_frontend.settings import COLOR_STYLE_CHOICES


class BadgeForm(EntangledModelForm):
    """
    Components > "Badge" Plugin
    https://getbootstrap.com/docs/5.0/components/badge/
    """

    class Meta:
        model = FrontendUIItem
        entangled_fields = {
            "config": [
                "badge_text",
                "badge_context",
                "badge_pills",
                "attributes",
            ]
        }

    badge_text = forms.CharField(
        label=_("Badge text"),
        max_length=255,
    )
    badge_context = forms.ChoiceField(
        label=_("Context"),
        choices=COLOR_STYLE_CHOICES,
        initial=first_choice(COLOR_STYLE_CHOICES),
        widget=ColoredButtonGroup(),
    )
    badge_pills = forms.BooleanField(
        label=_("Pills style"),
        initial=False,
        required=False,
        help_text=_("Activates the pills style."),
    )
    attributes = AttributesFormField()
    tag_type = TagTypeFormField()



from django.conf import settings
from django.utils.translation import gettext_lazy as _

TAB_TYPE_CHOICES = (
    ("nav-tabs", _("Tabs")),
    ("nav-pills", _("Pills")),
)

TAB_ALIGNMENT_CHOICES = (
    ("nav-fill", _("Fill")),
    ("nav-justified", _("Justified")),
    ("justify-content-start", _("Justify start")),
    ("justify-content-center", _("Justify center")),
    ("justify-content-end", _("Justify end")),
    ("flex-column", _("Column")),
)

TAB_EFFECT_CHOICES = getattr(
    settings,
    "DJANGOCMS_FRONTEND_TAB_EFFECTS",
    (("fade", _("Fade")),),
)

TAB_TEMPLATE_CHOICES = getattr(
    settings,
    "DJANGOCMS_FRONTEND_TAB_TEMPLATES",
    (("default", _("Default")),),
)



from ...helpers import get_related_object

try:
    from functools import cached_property
except ImportError:  # Only available since Python 3.8
    cached_property = property

from django.utils.translation import gettext_lazy as _
from django.utils.translation import ngettext

from djangocms_frontend.models import FrontendUIItem

from .constants import GRID_CONTAINER_CHOICES


class TitelModelMixin:
    pass


class GridContainer(TitelModelMixin, FrontendUIItem):
    """
    Layout > Grid: "Container" Plugin
    https://getbootstrap.com/docs/5.0/layout/grid/
    """

    class Meta:
        proxy = True
        verbose_name = _("Container")
        _("GridContainer")

    def get_short_description(self):
        text = self.config.get("plugin_title", {}).get("title", "") or self.config.get(
            "attributes", {}
        ).get("id", "")
        for item in GRID_CONTAINER_CHOICES[1:]:
            if item[0] == self.container_type:
                text += f" ({item[1]})"
        return text

    @cached_property
    def image(self):
        if getattr(self, "container_image", False):
            return get_related_object(self.config, "container_image")
        return None


class GridRow(TitelModelMixin, FrontendUIItem):
    """
    Layout > Grid: "Row" Plugin
    https://getbootstrap.com/docs/5.0/layout/grid/
    """

    class Meta:
        proxy = True
        verbose_name = _("Row")
        _("GridRow")

    def get_short_description(self):
        descr = self.config.get("plugin_title", {}).get("title", "") or self.config.get(
            "attributes", {}
        ).get("id", "")
        column_count = len(self.child_plugin_instances or [])
        column_count_str = ngettext(
            "(1 column)", "(%(count)i columns)", column_count
        ) % {"count": column_count}
        if descr:
            column_count_str = f"{descr} {column_count_str}"
        return column_count_str


class GridColumn(FrontendUIItem):
    """
    Layout > Grid: "Column" Plugin
    https://getbootstrap.com/docs/5.0/layout/grid/
    """

    class Meta:
        proxy = True
        verbose_name = _("Column")
        _("GridColumn")

    def get_short_description(self):
        text = self.config.get("plugin_title", {}).get("title", "") or self.config.get(
            "attributes", {}
        ).get("id", "")

        if self.xs_col:
            text += f" (col-{self.xs_col}) "
        else:
            text += "(auto) "
        return text.strip()



from django.utils.translation import gettext_lazy as _
from entangled.forms import EntangledModelFormMixin

from djangocms_frontend.fields import IconMultiselect, OptionalDeviceChoiceField
from djangocms_frontend.helpers import insert_fields
from djangocms_frontend.settings import DEVICE_CHOICES

visibility_class = {
    "GridRow": "flex",
}


def get_display_classes(visibility_set, visibility_class="block"):
    """Generates the necessary bootstrap display utility classes for a set
    of devices on which the content should be seen/hidden"""
    classes = []
    visible = True
    for device, __ in DEVICE_CHOICES:
        if (device in visibility_set) != visible:
            visible = device in visibility_set
            if device == "xs":
                classes.append(f"d-{visibility_class if visible else 'none'}")
            else:
                classes.append(f"d-{device}-{visibility_class if visible else 'none'}")
    return classes


class ResponsiveMixin:
    def get_fieldsets(self, request, obj=None):
        return insert_fields(
            super().get_fieldsets(request, obj),
            ("responsive_visibility",),
            block=None,
            position=-1,
            blockname=_("Visibility"),
        )

    def render(self, context, instance, placeholder):
        if instance.config.get("responsive_visibility", None) is not None:
            instance.add_classes(
                get_display_classes(
                    instance.responsive_visibility,
                    visibility_class.get(instance.ui_item, "block"),
                )
            )
        return super().render(context, instance, placeholder)


class ResponsiveFormMixin(EntangledModelFormMixin):
    class Meta:
        entangled_fields = {
            "config": [
                "responsive_visibility",
            ]
        }

    responsive_visibility = OptionalDeviceChoiceField(
        label=_("Show element on device"),
        required=False,
        help_text=_("Select only devices on which this element should be shown."),
        widget=IconMultiselect(),
    )



from cms.plugin_pool import plugin_pool
from django.utils.translation import gettext_lazy as _

from ... import settings
from ...cms_plugins import CMSUIPlugin
from ...common.attributes import AttributesMixin
from .. import badge
from . import forms, models

mixin_factory = settings.get_renderer(badge)


@plugin_pool.register_plugin
class BadgePlugin(mixin_factory("Badge"), AttributesMixin, CMSUIPlugin):
    """
    Components > "Badge" Plugin
    https://getbootstrap.com/docs/5.0/components/badge/
    """

    name = _("Badge")
    module = _("Frontend")
    model = models.Badge
    form = forms.BadgeForm
    render_template = "djangocms_frontend/badge.html"
    change_form_template = "djangocms_frontend/admin/badge.html"
    text_enabled = True

    fieldsets = [
        (
            None,
            {
                "fields": (
                    "badge_text",
                    "badge_context",
                    "badge_pills",
                )
            },
        ),
    ]



from copy import copy

from django import forms
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _
from entangled.forms import EntangledModelForm

from djangocms_frontend import settings
from djangocms_frontend.common.background import BackgroundFormMixin
from djangocms_frontend.common.responsive import ResponsiveFormMixin
from djangocms_frontend.common.sizing import SizingFormMixin
from djangocms_frontend.common.spacing import SpacingFormMixin
from djangocms_frontend.fields import (
    AttributesFormField,
    AutoNumberInput,
    ButtonGroup,
    IconGroup,
    TagTypeFormField,
)
from djangocms_frontend.helpers import first_choice, link_to_framework_doc
from djangocms_frontend.models import FrontendUIItem

from ...common.title import TitleFormMixin
from .. import grid
from .constants import (
    GRID_COLUMN_ALIGNMENT_CHOICES,
    GRID_CONTAINER_CHOICES,
    GRID_ROW_HORIZONTAL_ALIGNMENT_CHOICES,
    GRID_ROW_VERTICAL_ALIGNMENT_CHOICES,
    GRID_SIZE,
)

mixin_factory = settings.get_forms(grid)


class GridContainerForm(
    mixin_factory("GridContainer"),
    TitleFormMixin,
    BackgroundFormMixin,
    ResponsiveFormMixin,
    SpacingFormMixin,
    SizingFormMixin,
    EntangledModelForm,
):
    """
    Layout > Grid: "Container" Plugin
    https://getbootstrap.com/docs/5.0/layout/grid/
    """

    class Meta:
        model = FrontendUIItem
        entangled_fields = {
            "config": [
                "container_type",
                "attributes",
            ]
        }
        untangled_fields = ("tag_type",)

    container_type = forms.ChoiceField(
        label=_("Container type"),
        choices=GRID_CONTAINER_CHOICES,
        initial=first_choice(GRID_CONTAINER_CHOICES),
        help_text=_(
            "Defines if the grid should use fixed width, "
            "fluid width or the container should fill the full width without "
            "margins or padding."
        ),
    )
    attributes = AttributesFormField()
    tag_type = TagTypeFormField()


class GridRowBaseForm(
    mixin_factory("GridRow"),
    TitleFormMixin,
    ResponsiveFormMixin,
    SpacingFormMixin,
    EntangledModelForm,
):
    class Meta:
        model = FrontendUIItem
        entangled_fields = {
            "config": [
                "vertical_alignment",
                "horizontal_alignment",
                "gutters",
                "attributes",
            ]
        }
        untangled_fields = (
            "tag_type",
            "create",
        )

    create = forms.IntegerField(
        label=_("Create columns"),
        help_text=_("Number of columns to create when saving."),
        required=False,
        min_value=0,
        max_value=GRID_SIZE,
    )
    vertical_alignment = forms.ChoiceField(
        label=_("Vertical alignment"),
        choices=settings.EMPTY_CHOICE + GRID_ROW_VERTICAL_ALIGNMENT_CHOICES,
        required=False,
        help_text=link_to_framework_doc("GridRow", "vertical_alignment_link"),
        widget=IconGroup(),
    )
    horizontal_alignment = forms.ChoiceField(
        label=_("Horizontal alignment"),
        choices=settings.EMPTY_CHOICE + GRID_ROW_HORIZONTAL_ALIGNMENT_CHOICES,
        required=False,
        help_text=link_to_framework_doc("GridRow", "horizontal_alignment_link"),
        widget=IconGroup(),
    )
    gutters = forms.ChoiceField(
        label=_("Gutters"),
        initial=settings.EMPTY_CHOICE[0][0],
        choices=settings.EMPTY_CHOICE + settings.SPACER_SIZE_CHOICES,
        required=False,
        help_text=_("To remove all spaces between rows set gutters to 0."),
        widget=ButtonGroup(attrs=dict(property="text")),
    )
    attributes = AttributesFormField()
    tag_type = TagTypeFormField()


extra_fields_column = {}
for size in settings.DEVICE_SIZES:
    extra_fields_column[f"row_cols_{size}"] = forms.IntegerField(
        label="row-cols" if size == "xs" else f"row-cols-{size}",
        required=False,
        min_value=1,
        max_value=GRID_SIZE,
    )

GridRowForm = type(
    "GridRowBaseForm",
    (GridRowBaseForm,),
    copy(extra_fields_column),
)

GridRowForm.Meta.entangled_fields["config"] += extra_fields_column.keys()


class GridColumnBaseForm(
    mixin_factory("GridColumn"),
    TitleFormMixin,
    BackgroundFormMixin,
    ResponsiveFormMixin,
    SpacingFormMixin,
    EntangledModelForm,
):
    class Meta:
        model = FrontendUIItem
        entangled_fields = {
            "config": [
                "column_alignment",
                "text_alignment",
                "attributes",
            ]
        }
        untangled_fields = ("tag_type",)

    column_alignment = forms.ChoiceField(
        label=_("Column alignment"),
        choices=settings.EMPTY_CHOICE + GRID_COLUMN_ALIGNMENT_CHOICES,
        required=False,
        widget=IconGroup(),
    )
    text_alignment = forms.ChoiceField(
        label=_("Content alignment"),
        choices=settings.EMPTY_CHOICE + settings.ALIGN_CHOICES,
        required=False,
        widget=IconGroup(),
    )
    attributes = AttributesFormField()
    tag_type = TagTypeFormField()

    def clean(self):
        super().clean()
        for size in settings.DEVICE_SIZES:
            if f"{size}_col" in self.cleaned_data:
                if (
                    isinstance(self.cleaned_data[f"{size}_col"], str)
                    and self.cleaned_data[f"{size}_col"].isnumeric()
                ):
                    self.cleaned_data[f"{size}_col"] = int(
                        self.cleaned_data[f"{size}_col"]
                    )
            else:
                raise ValidationError(
                    _(
                        'Column size needs to be empty, "auto", or a '
                        "number between 1 and %(cols)d"
                    ),
                    params=dict(cols=GRID_SIZE),
                    code="invalid_column",
                )


# convert regular text type fields to number
# col_choices = [(col + 1, str(col + 1)) for col in range(GRID_SIZE)] + [("auto", "Auto")]
extra_fields_column = {}
for size in settings.DEVICE_SIZES:
    extra_fields_column[f"{size}_col"] = forms.IntegerField(
        label="col" if size == "xs" else f"col-{size}",
        required=False,
        initial="",
        min_value=0,
        max_value=GRID_SIZE,
        widget=AutoNumberInput(),
    )
    extra_fields_column[f"{size}_order"] = forms.IntegerField(
        label="order" if size == "xs" else f"order-{size}",
        required=False,
        min_value=0,
        max_value=GRID_SIZE,
        widget=forms.HiddenInput()
        if "{size}_order" in getattr(settings, "EXCL_COL_PROP", ())
        else forms.NumberInput(),
    )
    extra_fields_column[f"{size}_offset"] = forms.IntegerField(
        label="offset" if size == "xs" else f"offset-{size}",
        required=False,
        min_value=0,
        max_value=GRID_SIZE,
        widget=forms.HiddenInput()
        if "{size}_offset" in getattr(settings, "EXCL_COL_PROP", ())
        else forms.NumberInput(),
    )
    extra_fields_column[f"{size}_ms"] = forms.BooleanField(
        label="ms-auto" if size == "xs" else f"ms-{size}-auto",
        required=False,
        widget=forms.HiddenInput()
        if "{size}_ms" in getattr(settings, "EXCL_COL_PROP", ())
        else forms.CheckboxInput(),
    )
    extra_fields_column[f"{size}_me"] = forms.BooleanField(
        label="me-auto" if size == "xs" else f"me-{size}-auto",
        required=False,
        widget=forms.HiddenInput()
        if "{size}_me" in getattr(settings, "EXCL_COL_PROP", ())
        else forms.CheckboxInput(),
    )

GridColumnForm = type(
    "GridColumnForm",
    (GridColumnBaseForm,),
    copy(extra_fields_column),
)

GridColumnForm.Meta.entangled_fields["config"] += extra_fields_column.keys()



from cms.plugin_pool import plugin_pool
from django.utils.translation import gettext_lazy as _

from ... import settings
from ...cms_plugins import CMSUIPlugin
from ...common.attributes import AttributesMixin
from ...helpers import add_plugin
from .. import accordion
from . import forms, models

mixin_factory = settings.get_renderer(accordion)


@plugin_pool.register_plugin
class AccordionPlugin(mixin_factory("Accordion"), AttributesMixin, CMSUIPlugin):
    """
    Component > "Accordion" Plugin
    https://getbootstrap.com/docs/5.0/components/accordion/
    """

    name = _("Accordion")
    module = _("Frontend")
    model = models.Accordion
    form = forms.AccordionForm
    allow_children = True
    child_classes = [
        "AccordionItemPlugin",
    ]

    fieldsets = [
        (
            None,
            {
                "fields": (
                    "create",
                    (
                        "accordion_header_type",
                        "accordion_flush",
                    ),
                )
            },
        ),
    ]

    def save_model(self, request, obj, form, change):
        super().save_model(request, obj, form, change)
        data = form.cleaned_data
        for pos in range(data["create"] if data["create"] is not None else 0):
            item = models.AccordionItem(
                parent=obj,
                position=obj.position + 1 + pos,
                placeholder=obj.placeholder,
                language=obj.language,
                plugin_type=AccordionItemPlugin.__name__,
                ui_item=models.AccordionItem.__class__.__name__,
                config=dict(
                    accordion_item_header=_("Item {}").format(pos + 1),
                    accordion_item_open=(pos == 0),
                ),
            ).initialize_from_form(forms.AccordionItemForm)
            add_plugin(obj.placeholder, item)


@plugin_pool.register_plugin
class AccordionItemPlugin(mixin_factory("AccordionItem"), CMSUIPlugin):
    """
    Component > "Collapse" Plugin
    https://getbootstrap.com/docs/5.0/components/collapse/
    """

    name = _("Accordion item")
    module = _("Frontend")
    model = models.AccordionItem
    form = forms.AccordionItemForm
    allow_children = True
    parent_classes = [
        "AccordionPlugin",
    ]

    fieldsets = [
        (
            None,
            {
                "fields": (
                    "accordion_item_header",
                    "accordion_item_open",
                )
            },
        ),
        (
            _("Advanced settings"),
            {
                "classes": ("collapse",),
                "fields": (
                    "tag_type",
                    "attributes",
                ),
            },
        ),
    ]



# Generated by Django 3.2.11 on 2022-01-12 21:03

from django.db import migrations


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("djangocms_frontend", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="Collapse",
            fields=[],
            options={
                "proxy": True,
                "indexes": [],
                "constraints": [],
                "verbose_name": "Collapse",
            },
            bases=("djangocms_frontend.frontenduiitem",),
        ),
        migrations.CreateModel(
            name="CollapseContainer",
            fields=[],
            options={
                "proxy": True,
                "indexes": [],
                "constraints": [],
                "verbose_name": "Collapse container",
            },
            bases=("djangocms_frontend.frontenduiitem",),
        ),
        migrations.CreateModel(
            name="CollapseTrigger",
            fields=[],
            options={
                "proxy": True,
                "indexes": [],
                "constraints": [],
                "verbose_name": "Collapse trigger",
            },
            bases=("djangocms_frontend.frontenduiitem",),
        ),
    ]



from django.utils.translation import gettext_lazy as _

from djangocms_frontend.contrib.link.models import GetLinkMixin, Link
from djangocms_frontend.models import FrontendUIItem


class Navigation(FrontendUIItem):
    """
    Components > "Jumbotron" Plugin
    https://getbootstrap.com/docs/5.0/components/jumbotron/
    """

    class Meta:
        proxy = True
        verbose_name = _("Navigation")

    def get_short_description(self):
        return f'({self.config.get("navbar_design", "-")})'


class NavContainer(FrontendUIItem):
    class Meta:
        proxy = True
        verbose_name = _("Navigation container")


class NavLink(Link):
    class Meta:
        proxy = True
        verbose_name = _("Navigation Link")


class PageTree(FrontendUIItem):
    class Meta:
        proxy = True
        verbose_name = _("Page tree")


class NavBrand(GetLinkMixin, FrontendUIItem):
    class Meta:
        proxy = True
        verbose_name = _("Brand")



class BadgeRenderMixin:
    def render(self, context, instance, placeholder):
        instance.add_classes(f"badge bg-{instance.badge_context}")
        if instance.badge_pills:
            instance.add_classes("rounded-pill")
        return super().render(context, instance, placeholder)



class NavigationRenderMixin:
    def render(self, context, instance, placeholder):
        instance.add_classes(
            "navbar",
            f"navbar-{instance.config.get('navbar_design', '')}",
            f"navbar-expand-{instance.config.get('navbar_breakpoint', '')}",
        )
        return super().render(context, instance, placeholder)


class PageTreeRenderMixin:
    def render(self, context, instance, placeholder):
        instance.add_classes("navbar-nav")
        return super().render(context, instance, placeholder)


class NavBrandRenderMixin:
    def render(self, context, instance, placeholder):
        instance.add_classes("navbar-brand")
        return super().render(context, instance, placeholder)


class NavLinkRenderMixin:
    def render(self, context, instance, placeholder):
        instance.add_classes("nav-link")
        if instance.child_plugin_instances:
            instance.add_classes("dropdown-toggle")
        return super().render(context, instance, placeholder)



from django.conf import settings
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from easy_thumbnails.files import get_thumbnailer

from djangocms_frontend.contrib.link.models import GetLinkMixin
from djangocms_frontend.helpers import get_related_object
from djangocms_frontend.models import FrontendUIItem

# use golden ration as default (https://en.wikipedia.org/wiki/Golden_ratio)
PICTURE_RATIO = getattr(settings, "DJANGOCMS_PICTURE_RATIO", 1.6180)


class ImageMixin:
    image_field = None

    def get_size(self, width=None, height=None):
        crop = getattr(self, "use_crop", False)
        upscale = getattr(self, "use_upscale", False)
        # use field thumbnail settings
        if getattr(self, "thumbnail_options", None):
            thumbnail_options = get_related_object(self.config, "thumbnail_options")
            width = thumbnail_options.width
            height = thumbnail_options.height
            crop = thumbnail_options.crop
            upscale = thumbnail_options.upscale
        elif not getattr(self, "use_automatic_scaling", None):
            width = getattr(self, "width", None)
            height = getattr(self, "height", None)

        # calculate height when not given according to the
        # golden ratio or fallback to the image size
        if not height and width:
            height = width / PICTURE_RATIO
        elif not width and height:
            width = height * PICTURE_RATIO
        elif not width and not height and getattr(self, "picture", None):
            if self.rel_image:
                width = self.rel_image.width
                height = self.rel_image.height
            else:
                width = 0
                height = 0
        elif not width and not height:  # pragma: no cover
            # If no information is available on the image size whatsoever,
            # make it 640px wide and use PICTURE_RATIO
            width, height = 640, 640 / PICTURE_RATIO
        width = int(width)
        height = int(height)
        return {
            "size": (width, height),
            "crop": crop,
            "upscale": upscale,
        }

    @cached_property
    def rel_image(self):
        if self.config.get(self.image_field, None):
            return get_related_object(self.config, self.image_field)
        return None


class Image(GetLinkMixin, ImageMixin, FrontendUIItem):
    """
    Content > "Image" Plugin
    https://getbootstrap.com/docs/5.0/content/images/
    """

    class Meta:
        proxy = True
        verbose_name = _("Image")

    image_field = "picture"

    @property
    def is_responsive_image(self):
        if self.external_picture:
            return False
        if self.use_responsive_image == "inherit":
            return getattr(settings, "DJANGOCMS_PICTURE_RESPONSIVE_IMAGES", False)
        return self.use_responsive_image == "yes"

    @cached_property
    def img_srcset_data(self):
        if not (self.picture and self.is_responsive_image):
            return None

        srcset = []

        try:
            thumbnailer = get_thumbnailer(self.rel_image)

            picture_options = self.get_size(self.width, self.height)
            picture_width = picture_options["size"][0]
            thumbnail_options = {"crop": picture_options["crop"]}
            breakpoints = getattr(
                settings,
                "DJANGOCMS_PICTURE_RESPONSIVE_IMAGES_VIEWPORT_BREAKPOINTS",
                [576, 768, 992],
            )

            for size in filter(lambda x: x < picture_width, breakpoints):
                thumbnail_options["size"] = (size, size)
                srcset.append((int(size), thumbnailer.get_thumbnail(thumbnail_options)))
        except ValueError:
            # get_thumbnailer() raises this if it can't establish a `relative_name`.
            # This may mean that the filer image has been deleted
            pass

        return srcset

    @cached_property
    def img_src(self):
        # we want the external image to take priority by design
        # please open a ticket if you disagree for an open discussion
        if self.external_picture:
            return self.external_picture
        # image can be empty, for example when the image is removed from filer
        # in this case we want to return an empty string to avoid #69
        elif not self.picture:
            return ""
        # return the original, unmodified image
        elif self.use_no_cropping:
            return self.rel_image.url if self.rel_image else ""

        picture_options = self.get_size(
            width=self.width or 0,
            height=self.height or 0,
        )

        thumbnail_options = {
            "size": picture_options["size"],
            "crop": picture_options["crop"],
            "upscale": picture_options["upscale"],
            "subject_location": self.rel_image.subject_location
            if self.rel_image
            else (),
        }

        try:
            thumbnailer = get_thumbnailer(self.rel_image)
            url = thumbnailer.get_thumbnail(thumbnail_options).url
        except ValueError:
            # get_thumbnailer() raises this if it can't establish a `relative_name`.
            # This may mean that the filer image has been deleted
            url = ""
        return url

    def get_short_description(self):
        if self.external_picture:
            return self.external_picture
        if self.rel_image and self.rel_image.label:
            return self.rel_image.label
        return _("<file is missing>")



from django.test import TestCase

from djangocms_frontend.contrib.jumbotron.models import Jumbotron


class JumbotronModelTestCase(TestCase):
    def test_instance(self):
        instance = Jumbotron.objects.create()
        instance.config["jumbotron_fluid"] = False
        self.assertEqual(str(instance), "Jumbotron (1)")
        self.assertEqual(instance.get_short_description(), "")
        instance.config["jumbotron_fluid"] = True
        self.assertEqual(instance.get_short_description(), "(Fluid)")



from django.conf import settings
from django.utils.translation import gettext_lazy as _

ICON_CDN = {
    "bootstrap-icons": "https://cdn.jsdelivr.net/npm/bootstrap-icons@1.10.3/font/bootstrap-icons.css",
    "font-awesome": "https://cdnjs.cloudflare.com/ajax/libs/font-awesome/6.0.0/css/all.min.css",
    "material-icons-filled": "https://fonts.googleapis.com/css2?family=Material+Icons",
    "material-icons-outlined": "https://fonts.googleapis.com/css2?family=Material+Icons+Outlined",
    "material-icons-round": "https://fonts.googleapis.com/css2?family=Material+Icons+Round",
    "material-icons-sharp": "https://fonts.googleapis.com/css2?family=Material+Icons+Sharp",
    "material-icons-two-tone": "https://fonts.googleapis.com/css2?family=Material+Icons+Two+Tone",
    "fomantic-ui": "fomantic-ui-icons.css",
}

ICON_LIBRARIES_SHOWN = getattr(settings, "DJANGOCMS_FRONTEND_ICONS_LIBRARIES_SHOWN", (
    "font-awesome",
    "bootstrap-icons",
    "material-icons-filled",
    "material-icons-outlined",
    "material-icons-round",
    "material-icons-sharp",
    "material-icons-two-tone",
    "fomantic-ui",
    "foundation-icons",
    "elegant-icons",
    "feather-icons",
    "happy-icons",
    "icomoon",
    "open-iconic",
    "tabler-icons",
    "zondicons",
    "weather-icons",
))

ICON_LIBRARIES = getattr(
    settings,
    "DJANGOCMS_FRONTEND_ICON_LIBRARIES",
    {
        library: (f"{library}.min.json", ICON_CDN.get(library, f"{library}.css"))
        for library in getattr(
            settings,
            "DJANGOCMS_FRONTEND_ICON_LIBRARIES_SHOWN",
            ICON_LIBRARIES_SHOWN
        )
    },
)

ICON_SIZE_CHOICES = getattr(
    settings,
    "DJANGOCMS_FRONTEND_ICON_SIZE_CHOICES",
    (
        ("", _("Regular")),
        ("200%", _("x 2")),
        ("300%", _("x 3")),
        ("400%", _("x 4")),
        ("500%", _("x 5")),
        ("800%", _("x 8")),
        ("1200%", _("x 12")),
    ),
)

ICON_TAG_TYPES = getattr(
    settings,
    "DJANGOCMS_FRONTEND_ICON_TAG_TYPES",
    ("i", "span"),
)

ICON_TAG_TYPES = tuple((entry, entry) for entry in ICON_TAG_TYPES)



from django.utils.translation import gettext_lazy as _

from djangocms_frontend.helpers import insert_fields


class AttributesMixin:
    block_attr = {
        "description": _(
            "Advanced settings lets you add html attributes to render this element. Use them wisely and rarely."
        ),
        "classes": (
            "collapse",
            "attributes",
        ),
    }

    def get_fieldsets(self, request, obj=None):
        meta = self.form._meta
        fields = (
            ["tag_type"] if "tag_type" in getattr(meta, "untangled_fields", ()) else []
        )
        fields.append("attributes")
        return insert_fields(
            super().get_fieldsets(request, obj),
            fields,
            blockname=_("Advanced settings"),
            blockattrs=self.block_attr,
            position=-1,  # Always last
        )



from django import forms
from django.utils.translation import gettext_lazy as _
from entangled.forms import EntangledModelForm

from djangocms_frontend import settings
from djangocms_frontend.contrib import accordion
from djangocms_frontend.fields import AttributesFormField, TagTypeFormField
from djangocms_frontend.models import FrontendUIItem

mixin_factory = settings.get_forms(accordion)


class AccordionForm(mixin_factory("Accordion"), EntangledModelForm):
    """
    Components > "Accordion" Plugin
    https://getbootstrap.com/docs/5.0/components/accordion/
    """

    class Meta:
        model = FrontendUIItem
        entangled_fields = {
            "config": [
                "accordion_header_type",
                "accordion_flush",
                "attributes",
            ]
        }
        untangled_fields = (
            "tag_type",
            "create",
        )

    create = forms.IntegerField(
        label=_("Create accordion items"),
        help_text=_("Number of accordion items to create when saving."),
        required=False,
        initial=False,
        min_value=0,
        max_value=20,
    )
    accordion_header_type = forms.ChoiceField(
        label=_("Header type"),
        initial=settings.EMPTY_CHOICE[0][0],
        choices=settings.EMPTY_CHOICE + settings.HEADER_CHOICES,
        required=False,
    )
    accordion_flush = forms.BooleanField(
        label=_("Integrate into parent"),
        initial=False,
        required=False,
        help_text=_(
            "Removes the default background-color, some borders, and some rounded corners "
            "to render accordions edge-to-edge with their parent container "
        ),
    )
    attributes = AttributesFormField()
    tag_type = TagTypeFormField()


class AccordionItemForm(mixin_factory("AccordionItem"), EntangledModelForm):
    """
    Components > "AccordionItem" Plugin
    https://getbootstrap.com/docs/5.0/components/accordion/
    """

    class Meta:
        model = FrontendUIItem
        entangled_fields = {
            "config": [
                "accordion_item_header",
                "accordion_item_open",
                "attributes",
            ]
        }
        untangled_fields = ("tag_type",)

    accordion_item_header = forms.CharField(
        label=_("Header"),
        required=True,
    )
    accordion_item_open = forms.BooleanField(
        label=_("Item open"),
        initial=False,
        required=False,
        help_text=_("Initially shows this accordion item on page load."),
    )
    attributes = AttributesFormField()
    tag_type = TagTypeFormField()



# Generated by Django 3.2.11 on 2022-01-12 21:01

import django.core.serializers.json
import django.db.models.deletion
from cms.utils.compat import DJANGO_3_1
from django.db import migrations, models

import djangocms_frontend.fields

if DJANGO_3_1:
    from django_jsonfield_backport.models import JSONField
else:
    JSONField = models.JSONField


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("cms", "0022_auto_20180620_1551"),
    ]

    operations = [
        migrations.CreateModel(
            name="FrontendUIItem",
            fields=[
                (
                    "cmsplugin_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        related_name="%(app_label)s_%(class)s",
                        serialize=False,
                        to="cms.cmsplugin",
                    ),
                ),
                ("ui_item", models.CharField(max_length=30)),
                (
                    "tag_type",
                    djangocms_frontend.fields.TagTypeField(
                        blank=True,
                        default="div",
                        help_text="Select the HTML tag to be used.",
                        max_length=255,
                        verbose_name="Tag type",
                    ),
                ),
                (
                    "config",
                    JSONField(
                        default=dict,
                        encoder=django.core.serializers.json.DjangoJSONEncoder,
                    ),
                ),
            ],
            options={
                "verbose_name": "UI item",
            },
            bases=("cms.cmsplugin",),
        ),
    ]



from cms.plugin_pool import plugin_pool
from django.apps import apps
from django.conf import settings as django_settings
from django.utils.translation import gettext_lazy as _

from djangocms_frontend.helpers import get_plugin_template, insert_fields

from ... import settings
from ...cms_plugins import CMSUIPlugin
from ...common.attributes import AttributesMixin
from ...common.spacing import SpacingMixin
from .. import link
from . import forms, models
from .constants import USE_LINK_ICONS

mixin_factory = settings.get_renderer(link)


UILINK_FIELDS = (
    ("name", "link_type"),
    ("site", "url_grouper")
    if apps.is_installed("djangocms_url_manager")
    else ("external_link", "internal_link"),
    ("link_context", "link_size"),
    ("link_outline", "link_block"),
    "link_stretched",
)

UILINK_FIELDSET = [
    (
        None,
        {
            "fields": ("template",)
            + (
                UILINK_FIELDS + (("icon_left", "icon_right"),)
                if USE_LINK_ICONS
                else UILINK_FIELDS
            )
        },
    ),
]
if not apps.is_installed("djangocms_url_manager"):
    UILINK_FIELDSET += [
        (
            _("Link settings"),
            {
                "classes": ("collapse",),
                "fields": (
                    ("mailto", "phone"),
                    ("anchor", "target"),
                    ("file_link",),
                ),
            },
        ),
    ]


class LinkPluginMixin:
    link_fieldset_position = None
    link_fields = (
        (("site", "url_grouper"),)
        if apps.is_installed("djangocms_url_manager")
        else (
            ("external_link", "internal_link"),
            ("mailto", "phone"),
            ("anchor", "target"),
            "file_link",
        )
    )

    def get_form(self, request, obj=None, change=False, **kwargs):
        """The link form needs the request object to check permissions"""
        form = super().get_form(request, obj, change, **kwargs)
        form.request = request
        return form

    def get_fieldsets(self, request, obj=None):
        fieldsets = super().get_fieldsets(request, obj)
        if self.link_fieldset_position is not None:
            fieldsets = insert_fields(
                fieldsets,
                self.link_fields,
                blockname=_("Link settings"),
                position=self.link_fieldset_position,
            )
        return fieldsets


class LinkPlugin(
    mixin_factory("Link"), AttributesMixin, SpacingMixin, LinkPluginMixin, CMSUIPlugin
):
    """
    Components > "Button" Plugin
    https://getbootstrap.com/docs/5.0/components/buttons/
    """

    name = _("Link / Button")
    module = _("Frontend")
    model = models.Link
    form = forms.LinkForm
    change_form_template = "djangocms_frontend/admin/link.html"
    text_enabled = True
    allow_children = True

    fieldsets = UILINK_FIELDSET

    def get_render_template(self, context, instance, placeholder):
        return get_plugin_template(
            instance, "link", "link", settings.LINK_TEMPLATE_CHOICES
        )


if "djangocms_frontend.contrib.link" in django_settings.INSTALLED_APPS and 'LinkPlugin' not in plugin_pool.plugins:
    #  Only register plugin if in INSTALLED_APPS

    plugin_pool.register_plugin(LinkPlugin)



from django.utils.translation import gettext_lazy as _

CARD_LAYOUT_TYPE_CHOICES = (
    ("card-group", _("Card group")),
    ("row", _("Grid cards")),  # Removed in 5
)

CARD_ALIGNMENT_CHOICES = (
    ("start", _("Left")),
    ("center", _("Center")),
    ("end", _("Right")),
)

CARD_INNER_TYPE_CHOICES = (
    ("card-body", _("Body")),
    ("card-header", _("Header")),
    ("card-footer", _("Footer")),
    ("card-img-overlay", _("Image overlay")),
)

CARD_TAG_CHOICES = (
    ("div", "DIV"),
    ("h1", "H1"),
    ("h2", "H2"),
    ("h3", "H3"),
    ("h4", "H4"),
    ("h5", "H5"),
    ("h6", "H6"),
    ("p", "P"),
    ("small", "SMALL"),
)



from django.utils.translation import gettext_lazy as _

from djangocms_frontend.models import FrontendUIItem


class Badge(FrontendUIItem):
    """
    Components > "Badge" Plugin
    https://getbootstrap.com/docs/5.0/components/badge/
    """

    class Meta:
        proxy = True
        verbose_name = _("Badge")

    def get_short_description(self):
        return f"({self.badge_context})"



from django.contrib.auth.mixins import LoginRequiredMixin
from django.http import JsonResponse
from django.views.generic import View

from .helpers import get_link_choices


class AutocompleteJsonView(LoginRequiredMixin, View):
    """Handle AutocompleteWidget's AJAX requests for data."""

    paginate_by = 20
    admin_site = None

    def get(self, request, *args, **kwargs):
        """
        Return a JsonResponse with search results of the form:
        {
            results: [{id: "123" text: "foo"}],
            pagination: {more: true}
        }
        """

        # TODO Check permissions
        # ======================
        self.term = kwargs.get("term", request.GET.get("term", ""))
        results = get_link_choices(request, self.term)
        return JsonResponse(
            {
                "results": results,
                "pagination": {"more": False},
            }
        )



from django.test import TestCase

from djangocms_frontend.contrib.grid.forms import GridColumnForm, GridContainerForm
from djangocms_frontend.contrib.grid.models import GridColumn, GridContainer, GridRow


class GridModelTestCase(TestCase):
    def test_grid_instance(self):
        instance = GridContainer.objects.create()
        instance.initialize_from_form(GridContainerForm)
        self.assertEqual(str(instance), "GridContainer (1)")
        self.assertEqual(instance.get_short_description(), "")
        instance.config["plugin_title"] = {"show": False, "title": "test container"}
        self.assertEqual(instance.get_short_description(), "test container")

    def test_row_instance(self):
        instance = GridRow.objects.create()
        self.assertEqual(str(instance), "GridRow (1)")
        self.assertEqual(instance.get_short_description(), "(0 columns)")

    def test_column_instance(self):
        instance = GridColumn.objects.create()
        instance.initialize_from_form(GridColumnForm)
        self.assertEqual(str(instance), "GridColumn (1)")
        self.assertEqual(instance.get_short_description(), "(auto)")
        instance.config["xs_col"] = 12
        self.assertEqual(
            instance.get_short_description(),
            "(col-12)",
        )
        instance.config["md_col"] = 12
        instance.config["md_offset"] = 12
        instance.config["xs_offset"] = 12
        self.assertEqual(
            instance.get_short_description(),
            "(col-12)",
        )
        instance.config["xs_ms"] = 12
        instance.config["md_ms"] = 12
        self.assertEqual(
            instance.get_short_description(),
            "(col-12)",
        )



from django import apps, forms
from django.conf import settings as django_settings
from django.contrib.admin.widgets import SELECT2_TRANSLATIONS
from django.contrib.contenttypes.models import ContentType
from django.contrib.sites.models import Site
from django.core.exceptions import ObjectDoesNotExist, ValidationError
from django.db import models
from django.db.models.fields.related import ManyToOneRel
from django.utils.encoding import force_str
from django.utils.translation import get_language
from django.utils.translation import gettext as _
from django_select2.forms import HeavySelect2Widget, Select2Widget

# from djangocms_link.validators import IntranetURLValidator
from entangled.forms import EntangledModelForm
from filer.fields.image import AdminFileFormField, FilerFileField
from filer.models import File

from ... import settings
from ...common.spacing import SpacingFormMixin
from ...fields import (
    AttributesFormField,
    ButtonGroup,
    ColoredButtonGroup,
    TagTypeFormField,
    TemplateChoiceMixin,
)
from ...helpers import first_choice, get_related_object
from ...models import FrontendUIItem
from .. import link
from .constants import LINK_CHOICES, LINK_SIZE_CHOICES, TARGET_CHOICES
from .helpers import get_choices, get_object_for_value

mixin_factory = settings.get_forms(link)

if "djangocms_frontend.contrib.icon" in django_settings.INSTALLED_APPS:
    # Weak dependency on djangocms_frontend.contrib.icon
    from djangocms_frontend.contrib.icon.fields import IconPickerField
elif "djangocms_icon" in django_settings.INSTALLED_APPS:  # pragma: no cover
    # Weak dependency on djangocms_icon
    # (Even if djangocms_icon is in the python path, the admin form will fail due to missing
    # templates if it's not in INSTALLED_APPS)
    from djangocms_icon.fields import IconField as IconPickerField
else:  # pragma: no cover

    class IconPickerField(forms.CharField):  # lgtm [py/missing-call-to-init]
        def __init__(self, *args, **kwargs):
            kwargs["widget"] = forms.HiddenInput
            super().__init__(*args, **kwargs)


HOSTNAME = getattr(settings, "DJANGOCMS_LINK_INTRANET_HOSTNAME_PATTERN", None)
LINK_MODELS = getattr(django_settings, "DJANGOCMS_FRONTEND_LINK_MODELS", [])
MINIMUM_INPUT_LENGTH = getattr(
    django_settings, "DJANGOCMS_FRONTEND_MINIMUM_INPUT_LENGTH", 0
)


class Select2jqWidget(HeavySelect2Widget if MINIMUM_INPUT_LENGTH else Select2Widget):
    """Make jQuery available to Select2 widget"""

    empty_label = _("Select a destination")

    @property
    def media(self):
        extra = ".min"
        i18n_name = SELECT2_TRANSLATIONS.get(get_language())
        i18n_file = (
            ("admin/js/vendor/select2/i18n/%s.js" % i18n_name,) if i18n_name else ()
        )
        return forms.Media(
            js=("admin/js/vendor/select2/select2.full%s.js" % extra,)
            + i18n_file
            + ("djangocms_frontend/js/django_select2.js",),
            css={
                "screen": (
                    "admin/css/vendor/select2/select2%s.css" % extra,
                    "djangocms_frontend/css/select2.css",
                ),
            },
        )

    def __init__(self, *args, **kwargs):
        if MINIMUM_INPUT_LENGTH:
            if "attrs" in kwargs:
                kwargs["attrs"].setdefault(
                    "data-minimum-input-length", MINIMUM_INPUT_LENGTH
                )
            else:
                kwargs["attrs"] = {"data-minimum-input-length": MINIMUM_INPUT_LENGTH}
            kwargs.setdefault("data_view", "dcf_autocomplete:ac_view")
        super().__init__(*args, **kwargs)


class SmartLinkField(forms.ChoiceField):
    widget = Select2jqWidget()

    def prepare_value(self, value):
        if value:
            if isinstance(value, dict):  # Entangled dictionary?
                try:
                    app_label, model = value["model"].rsplit(".", 1)
                    content_type = ContentType.objects.get(
                        app_label=app_label, model=model
                    )
                    return f"{content_type.id}-{value['pk']}"
                except (TypeError, ValueError, KeyError, ObjectDoesNotExist):
                    pass
            elif isinstance(value, models.Model):
                content_type = ContentType.objects.get_for_model(value)
                return f"{content_type.id}-{value.id}"
        return ""

    def clean(self, value):
        obj = get_object_for_value(value)
        if obj is not None:
            return obj
        return super().clean(value)


if apps.apps.is_installed("djangocms_url_manager"):
    from djangocms_url_manager.forms import (
        HtmlLinkSiteSelectWidget,
        HtmlLinkUrlSelectWidget,
    )
    from djangocms_url_manager.models import UrlGrouper

    class AbstractLinkForm(EntangledModelForm):
        class Meta:
            entangled_fields = {
                "config": [
                    "site",
                    "url_grouper",
                ]
            }

        site = forms.ModelChoiceField(
            label=_("Site"),
            queryset=Site.objects.all(),
            widget=HtmlLinkSiteSelectWidget(
                attrs={"data-placeholder": _("Select site")}
            ),
            required=False,
        )
        url_grouper = forms.ModelChoiceField(
            label=_("Url"),
            queryset=UrlGrouper.objects.all(),
            widget=HtmlLinkUrlSelectWidget(
                attrs={"data-placeholder": _("Select URL object from list")}
            ),
            required=False,
        )

else:

    class AbstractLinkForm(EntangledModelForm):
        class Meta:
            entangled_fields = {
                "config": [
                    "external_link",
                    "internal_link",
                    "file_link",
                    "anchor",
                    "mailto",
                    "phone",
                    "target",
                ]
            }

        link_is_optional = False

        # url_validators = [
        #     IntranetURLValidator(intranet_host_re=HOSTNAME),
        # ]

        external_link = forms.URLField(
            label=_("External link"),
            required=False,
            #        validators=url_validators,
            help_text=_("Provide a link to an external source."),
        )
        internal_link = SmartLinkField(
            label=_("Internal link"),
            required=False,
            help_text=_("If provided, overrides the external link."),
        )
        file_link = AdminFileFormField(
            rel=ManyToOneRel(FilerFileField, File, "id"),
            queryset=File.objects.all(),
            to_field_name="id",
            label=_("File link"),
            required=False,
            help_text=_("If provided links a file from the filer app."),
        )
        # other link types
        anchor = forms.CharField(
            label=_("Anchor"),
            required=False,
            help_text=_(
                "Appends the value only after the internal or external link. "
                'Do <em>not</em> include a preceding "&#35;" symbol.'
            ),
        )
        mailto = forms.EmailField(
            label=_("Email address"),
            required=False,
        )
        phone = forms.CharField(
            label=_("Phone"),
            required=False,
        )
        # advanced options
        target = forms.ChoiceField(
            label=_("Target"),
            choices=settings.EMPTY_CHOICE + TARGET_CHOICES,
            required=False,
        )

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.fields["internal_link"].choices = self.get_choices

        def get_choices(self):
            if MINIMUM_INPUT_LENGTH == 0:
                return get_choices(self.request)
            if not self.is_bound:  # find inital value
                int_link_field = self.fields["internal_link"]
                initial = self.get_initial_for_field(int_link_field, "internal_link")
                if initial:  # Initial set?
                    obj = get_related_object(dict(obj=initial), "obj")  # get it!
                    if obj is not None:
                        value = int_link_field.prepare_value(initial)
                        return ((value, str(obj)),)
            return ()  # nothing found

        def clean(self):
            super().clean()
            link_field_names = (
                "external_link",
                "internal_link",
                "mailto",
                "phone",
                "file_link",
            )
            anchor_field_name = "anchor"
            field_names_allowed_with_anchor = (
                "external_link",
                "internal_link",
            )
            anchor_field_verbose_name = force_str(self.fields[anchor_field_name].label)
            anchor_field_value = self.cleaned_data.get(anchor_field_name, None)
            link_fields = {
                key: self.cleaned_data.get(key, None) for key in link_field_names
            }
            link_field_verbose_names = {
                key: force_str(self.fields[key].label) for key in link_fields.keys()
            }
            provided_link_fields = {
                key: value for key, value in link_fields.items() if value
            }

            if len(provided_link_fields) > 1:
                # Too many fields have a value.
                verbose_names = sorted(link_field_verbose_names.values())
                error_msg = _("Only one of {0} or {1} may be given.").format(
                    ", ".join(verbose_names[:-1]),
                    verbose_names[-1],
                )
                errors = {}.fromkeys(provided_link_fields.keys(), error_msg)
                raise ValidationError(errors)

            if (
                len(provided_link_fields) == 0
                and not self.cleaned_data.get(anchor_field_name, None)
                and not self.link_is_optional
            ):
                raise ValidationError(_("Please provide a link."))

            if anchor_field_value:
                for field_name in provided_link_fields.keys():
                    if field_name not in field_names_allowed_with_anchor:
                        error_msg = _(
                            "%(anchor_field_verbose_name)s is not allowed together with %(field_name)s"
                        ) % {
                            "anchor_field_verbose_name": anchor_field_verbose_name,
                            "field_name": link_field_verbose_names.get(field_name),
                        }
                        raise ValidationError(
                            {
                                anchor_field_name: error_msg,
                                field_name: error_msg,
                            }
                        )


class LinkForm(
    mixin_factory("Link"), SpacingFormMixin, TemplateChoiceMixin, AbstractLinkForm
):
    class Meta:
        model = FrontendUIItem
        entangled_fields = {
            "config": [
                "name",
                "template",
                "link_type",
                "link_context",
                "link_size",
                "link_outline",
                "link_block",
                "icon_left",
                "icon_right",
                "link_stretched",
                "attributes",
            ]
        }
        untangled_fields = ()

    name = forms.CharField(
        label=_("Display name"),
        required=False,
    )
    template = forms.ChoiceField(
        label=_("Template"),
        choices=settings.LINK_TEMPLATE_CHOICES,
        initial=first_choice(settings.LINK_TEMPLATE_CHOICES),
    )
    link_stretched = forms.BooleanField(
        label=_("Stretch link"),
        required=False,
        initial=False,
        help_text=_(
            "Stretches the active link area to the containing block (with position: relative)."
        ),
    )
    link_type = forms.ChoiceField(
        label=_("Type"),
        choices=LINK_CHOICES,
        initial=first_choice(LINK_CHOICES),
        widget=forms.RadioSelect(attrs={"class": "inline-block"}),
        help_text=_("Adds either a text link or a button which links to the target."),
    )
    link_context = forms.ChoiceField(
        label=_("Context"),
        choices=settings.EMPTY_CHOICE + settings.COLOR_STYLE_CHOICES,
        initial=settings.EMPTY_CHOICE[0][0],
        required=False,
        widget=ColoredButtonGroup(),
    )
    link_size = forms.ChoiceField(
        label=_("Button size"),
        choices=LINK_SIZE_CHOICES,
        initial=LINK_SIZE_CHOICES[1][0],  # Medium
        required=False,
        widget=ButtonGroup(
            attrs=dict(property="link-size", label_class="btn-secondary")
        ),
    )
    link_outline = forms.BooleanField(
        label=_("Outline"),
        initial=False,
        required=False,
        help_text=_("Removes the coloring from a button and keeps the outline."),
    )
    link_block = forms.BooleanField(
        label=_("Block"),
        initial=False,
        required=False,
        help_text=_("Extends the button to the width of its container."),
    )
    icon_left = IconPickerField(
        label=_("Icon left"),
        initial="",
        required=False,
    )
    icon_right = IconPickerField(
        label=_("Icon right"),
        initial="",
        required=False,
    )
    attributes = AttributesFormField()
    tag_type = TagTypeFormField()



from django.contrib.contenttypes.models import ContentType

plugin_migrations = {
    "djangocms_styledlink.StyledLink -> link.Link": [
        "(default) -> template",
        "label -> name",
        "title",
        "ext_destination -> external_link",
        "ext_follow",
        "mailto",
        "target -> link_target",
        "page_destination -> anchor",
        "() -> phone",
        "() -> link_context",
        "() -> link_size",
        "() -> link_outline",
        "() -> link_block",
        "() -> icon_left",
        "() -> icon_right",
        "() -> file_link",
        "S001",
    ],
    "cms_plugins.ImageContainerModel -> grid.GridContainer": [
        "image -> container_image",
        "(container-fluid) -> container_type",
        "positioning",
        "S002",
    ],
}


def s001_migrate_styled_link(obj, new_obj):
    if obj.int_destination_type_id:
        content_type = ContentType.objects.get(id=obj.int_destination_type_id)
        new_obj.config["internal_link"] = dict(
            model=f"{content_type.app_label}.{content_type.model}",
            pk=obj.int_destination_id,
        )
        styles = obj.styles.all()
        new_obj.config["attributes"] = {
            "class": " ".join(style.link_class for style in styles)
        }
        for style in styles:
            obj.styles.remove(style)
        new_obj.config["link_type"] = (
            "btn" if "btn" in new_obj.attributes["class"] else "link"
        )


def s002_migrate_image_container(obj, new_obj):
    new_obj.config["attributes"] = {}
    if obj.additional_classes:
        new_obj.config["attributes"]["class"] = obj.additional_classes
    if obj.additional_styles:
        new_obj.config["attributes"]["style"] = obj.additional_styles

    if "rgba(" in obj.color:
        _, color = obj.color.replace(" ", "").replace(")", "").split("(", 1)
        colors = color.split(",")
        colors[-1].replace(")", "")
        try:
            r, g, b, a = map(float, colors)
            new_obj.config["container_opaqueness"] = int(100 * a)
            if r > 250 and g > 250 and b > 250:
                new_obj.config["container_context"] = "light"
            if r < 5 and g < 5 and b < 5:
                new_obj.config["container_context"] = "dark"
        except (TypeError, AttributeError, ValueError):
            print("Could not convert color", obj.color)
    if obj.blur:
        if "px" in obj.blur:
            try:
                blur = int(obj.blur.replace("px", ""))
                new_obj.config["container_blur"] = blur
            except TypeError:
                print("Could not convert blur", obj.blur)
    if obj.positioning:
        items = obj.positioning.split(";")
        key, value = items[0].split(":", 1)
        new_obj.config["image_position"] = " ".join(value.split()[:2])


data_migration = {
    "S001": s001_migrate_styled_link,
    "S002": s002_migrate_image_container,
}

plugin_prefix = "StyledLink"



from cms.plugin_pool import plugin_pool
from django.utils.translation import gettext_lazy as _

from ... import settings
from ...cms_plugins import CMSUIPlugin
from ...common.attributes import AttributesMixin
from ...common.background import BackgroundMixin
from ...helpers import first_choice, get_plugin_template, get_template_path
from .. import navigation
from ..link.cms_plugins import LinkPlugin, LinkPluginMixin
from . import forms, models

mixin_factory = settings.get_renderer(navigation)

default_template = first_choice(settings.NAVIGATION_TEMPLATE_CHOICES)


@plugin_pool.register_plugin
class NavigationPlugin(
    mixin_factory("Navigation"),
    AttributesMixin,
    BackgroundMixin,
    CMSUIPlugin,
):
    """
    Creates a Navbar
    """

    name = _("Navigation")
    module = _("Frontend")
    model = models.Navigation
    form = forms.NavigationForm
    change_form_template = "djangocms_frontend/admin/navigation.html"
    allow_children = True
    child_classes = [
        "NavLinkPlugin",
        "PageTreePlugin",
        "NavBrandPlugin",
        "NavContainerPlugin",
    ]

    fieldsets = [
        (
            None,
            {
                "fields": (
                    "template",
                    (
                        "navbar_design",
                        "navbar_breakpoint",
                    ),
                    "navbar_container",
                )
            },
        ),
    ]

    def get_render_template(self, context, instance, placeholder):
        return get_plugin_template(
            instance, "navigation", "navigation", settings.NAVIGATION_TEMPLATE_CHOICES
        )

    def render(self, context, instance, placeholder):
        context["nav_template"] = instance.config.get("template", default_template)
        return super().render(context, instance, placeholder)


@plugin_pool.register_plugin
class PageTreePlugin(
    mixin_factory("PageTree"),
    AttributesMixin,
    CMSUIPlugin,
):
    name = _("Page tree")
    module = _("Frontend")
    model = models.PageTree
    form = forms.PageTreeForm
    change_form_template = "djangocms_frontend/admin/page_tree.html"
    allow_children = False
    parent_classes = ["NavigationPlugin", "NavContainerPlugin"]
    fieldsets = [
        (
            None,
            {"fields": ("start_level",)},
        ),
    ]

    def get_render_template(self, context, instance, placeholder):
        return get_template_path(
            "navigation", context.get("nav_template", default_template), "page_tree"
        )

    def render(self, context, instance, placeholder):
        context["start_level"] = instance.config.get("start_level", 0)
        return super().render(context, instance, placeholder)


@plugin_pool.register_plugin
class NavBrandPlugin(
    mixin_factory("NavBrand"),
    AttributesMixin,
    LinkPluginMixin,
    CMSUIPlugin,
):
    name = _("Brand")
    module = _("Frontend")
    model = models.NavBrand
    form = forms.NavBrandForm
    change_form_template = "djangocms_frontend/admin/brand.html"
    allow_children = True
    parent_classes = ["NavigationPlugin", "NavContainerPlugin"]
    link_fieldset_position = -1

    fieldsets = [
        (
            None,
            {"fields": ("simple_content",)},
        ),
    ]

    def get_render_template(self, context, instance, placeholder):
        return get_template_path(
            "navigation", context.get("nav_template", default_template), "brand"
        )


@plugin_pool.register_plugin
class NavContainerPlugin(
    mixin_factory("NavLink"),
    AttributesMixin,
    CMSUIPlugin,
):
    name = _("Navigation container")
    module = _("Frontend")
    model = models.NavContainer
    form = forms.NavContainerForm
    change_form_template = "djangocms_frontend/admin/nav_container.html"
    allow_children = True
    parent_classes = ["NavigationPlugin"]
    child_classes = [
        "NavLinkPlugin",
        "PageTreePlugin",
        "NavBrandPlugin",
    ]

    fieldsets = [
        (
            None,
            {"fields": ()},
        ),
    ]

    def get_render_template(self, context, instance, placeholder):
        return get_template_path(
            "navigation", context.get("nav_template", default_template), "nav_container"
        )


@plugin_pool.register_plugin
class NavLinkPlugin(
    mixin_factory("NavLink"),
    LinkPlugin,
):
    name = _("Navigation link")
    module = _("Frontend")
    model = models.NavLink
    form = forms.NavLinkForm
    change_form_template = "djangocms_frontend/admin/navlink.html"
    allow_children = True
    parent_classes = ["NavigationPlugin", "NavContainerPlugin", "NavLinkPlugin"]
    child_classes = [
        "NavLinkPlugin",
        "GridContainerPlugin",
        "GridRowPlugin",
        "ListGroupPlugin",
    ]

    def get_render_template(self, context, instance, placeholder):
        return get_template_path(
            "navigation", context.get("nav_template", default_template), "link"
        )



from django.test import TestCase

from djangocms_frontend.contrib.content.forms import (
    BlockquoteForm,
    CodeForm,
    FigureForm,
)
from djangocms_frontend.contrib.content.models import Blockquote, CodeBlock, Figure


class CodeBlockForm:
    pass


class ContentModelTestCase(TestCase):
    def test_code_instance(self):
        instance = CodeBlock.objects.create()
        instance.initialize_from_form(CodeForm).save()
        self.assertEqual(str(instance), "CodeBlock (1)")
        self.assertEqual(instance.get_short_description(), "<code>")

    def test_blockquote_instance(self):
        instance = Blockquote.objects.create()
        instance.initialize_from_form(BlockquoteForm).save()
        self.assertEqual(str(instance), "Blockquote (1)")
        self.assertEqual(instance.get_short_description(), "")

    def test_figure_instance(self):
        instance = Figure.objects.create()
        instance.initialize_from_form(FigureForm).save()
        self.assertEqual(str(instance), "Figure (1)")
        self.assertEqual(instance.get_short_description(), "")



from django.test import TestCase

from djangocms_frontend.contrib.listgroup.forms import ListGroupForm, ListGroupItemForm
from djangocms_frontend.contrib.listgroup.models import ListGroup, ListGroupItem


class ListGroupModelTestCase(TestCase):
    def test_list_group_instance(self):
        instance = ListGroup.objects.create().initialize_from_form(ListGroupForm)
        self.assertEqual(str(instance), "ListGroup (1)")
        self.assertEqual(instance.get_short_description(), "")
        instance.list_group_flush = True
        self.assertEqual(instance.get_short_description(), ".list-group-flush")

    def test_list_group_item_instance(self):
        instance = ListGroupItem.objects.create().initialize_from_form(
            ListGroupItemForm
        )
        self.assertEqual(str(instance), "ListGroupItem (1)")
        self.assertEqual(instance.get_short_description(), "")
        instance.list_context = "primary"
        self.assertEqual(instance.get_short_description(), ".list-group-item-primary")
        instance.list_state = "active"
        self.assertEqual(
            instance.get_short_description(), ".list-group-item-primary .active"
        )



from django import forms
from django.utils.translation import gettext_lazy as _
from entangled.forms import EntangledModelFormMixin

from djangocms_frontend import settings
from djangocms_frontend.fields import ButtonGroup
from djangocms_frontend.helpers import insert_fields


class AlertRenderMixin:
    render_template = "djangocms_frontend/bootstrap5/alert.html"

    def render(self, context, instance, placeholder):
        instance.add_classes(f"alert alert-{instance.alert_context}")
        if getattr(instance, "alert_shadow", ""):
            if instance.alert_shadow == "reg":
                instance.add_classes("shadow")
            else:
                instance.add_classes(f"shadow-{instance.alert_shadow}")
        if instance.alert_dismissible:
            instance.add_classes("alert-dismissible")
        return super().render(context, instance, placeholder)

    def get_fieldsets(self, request, obj=None):
        return insert_fields(
            super().get_fieldsets(request, obj),
            ("alert_shadow",),
            block=0,
            position=1,
        )


class AlertFormMixin(EntangledModelFormMixin):
    class Meta:
        entangled_fields = {
            "config": [
                "alert_shadow",
            ]
        }

    alert_shadow = forms.ChoiceField(
        label=_("Shadow"),
        required=False,
        choices=settings.EMPTY_CHOICE + settings.framework_settings.SHADOW_CHOICES,
        initial=settings.EMPTY_CHOICE[0][0],
        widget=ButtonGroup(attrs=dict(property="shadow")),
        help_text=_("Use shadows to optically lift alerts from the background."),
    )



from djangocms_frontend import settings


def m002_responsive_mixin(obj, new_obj):
    classes = new_obj.config["attributes"].get("class", "").split()
    if classes:
        display = (
            "block",
            "flex",
        )
        hidden = "none"

        visible = True
        hit = False
        responsive = []

        for device, _ in settings.DEVICE_CHOICES:
            stump = f"d-{device}-" if device != "xs" else "d-"
            if f"{stump}{hidden}" in classes and visible:
                visible = False
                hit = True
                classes.remove(f"{stump}{hidden}")
            for type in display:
                if f"{stump}{type}" in classes and not visible:
                    visible = True
                    hit = True
                    classes.remove(f"{stump}{type}")
            if visible:
                responsive.append(device)
        if hit:
            new_obj.config["responsive_visibility"] = responsive
            if classes:
                new_obj.config["attributes"]["class"] = " ".join(classes)
            else:
                new_obj.config["attributes"].pop("class")
        else:
            new_obj.config["responsive_visibility"] = None


def m003_background_mixin(obj, new_obj):
    classes = new_obj.config["attributes"].get("class", "").split()
    if classes:
        for context, _ in settings.COLOR_STYLE_CHOICES:
            if f"bg-{context}" in classes:
                new_obj.config["background_context"] = context
                classes.remove(f"bg-{context}")
        for cls, key in {
            "shadow-none": "none",
            "shadow-sm": "sm",
            "shadow": "reg",
            "shadow-lg": "lg",
        }.items():
            if cls in classes:
                new_obj.config["background_shadow"] = key
                classes.remove(cls)
        if classes:
            new_obj.config["attributes"]["class"] = " ".join(classes)
        else:
            new_obj.config["attributes"].pop("class")


def i001_icon(obj, new_obj):
    """Convert icons (only works for fontawesome)"""
    classes = obj.icon.split()

    # Translate from fontawesome 5 to fontawesome 6
    if "fas" in classes:
        library = "fa-solid"
        classes.pop(classes.index("fas"))
    elif "far" in obj.icon:
        library = "fa-regular"
        classes.pop(classes.index("far"))
    elif "fab" in obj.icon:
        library = "fa-brands"
        classes.pop(classes.index("fab"))
    else:
        # Not recognized. Keep classes. Icon, however, probably not visible in the admin.
        new_obj.config["icon"] = {
            "iconClass": obj.icon,
        }
        return

    new_obj.config["icon"] = {
        "libraryId": library,
        "libraryName": "fontAwesome",
        "iconHtml": f'<i class="{library} {" ".join(classes)}"></i>',
        "iconMarkup": "&",
        "iconClass": f'{library} {" ".join(classes)}',
        "iconText": "",
        "library": "font-awesome",
    }


def m001_spacing_mixin(obj, new_obj, type):
    classes = new_obj.config["attributes"].get("class", "").split()
    if classes:
        for size, _ in list(settings.SPACER_SIZE_CHOICES) + (
            [("auto", "auto")] if type == "margin" else []
        ):
            if f"{type[0]}-{size}" in classes:
                classes.remove(f"{type[0]}-{size}")
                classes.append(f"{type[0]}x-{size}")
                classes.append(f"{type[0]}y-{size}")
            for side, _ in settings.SPACER_X_SIDES_CHOICES:
                if f"{type[0]}{side}-{size}" in classes and not new_obj.config.get(
                    f"{type}_x", None
                ):
                    new_obj.config[f"{type}_x"] = f"{type[0]}{side}-{size}"
                    new_obj.config[f"{type}_devices"] = None
                    classes.remove(f"{type[0]}{side}-{size}")
            for side, _ in settings.SPACER_Y_SIDES_CHOICES:
                if f"{type[0]}{side}-{size}" in classes and not new_obj.config.get(
                    f"{type}_y", None
                ):
                    new_obj.config[f"{type}_y"] = f"{type[0]}{side}-{size}"
                    new_obj.config[f"{type}_devices"] = None
                    classes.remove(f"{type[0]}{side}-{size}")
        if classes:
            new_obj.config["attributes"]["class"] = " ".join(classes)
        else:
            new_obj.config["attributes"].pop("class")


plugin_migration = {
    "djangocms_icon.Icon -> icon.Icon": [
        "attributes",
        "I001_ICON",
        "M001-m",
        "M001-p",
        "M002",
        "M003",
    ],
}
data_migration = {
    "I001_ICON": i001_icon,
    "M001-m": lambda x, y: m001_spacing_mixin(x, y, "margin"),
    "M001-p": lambda x, y: m001_spacing_mixin(x, y, "padding"),
    "M002": m002_responsive_mixin,
    "M003": m003_background_mixin,
}

plugin_prefix = "djangocms_icon"



import importlib

from django.apps import apps
from django.conf import settings
from django.db import connection, models

from .base import SubcommandsCommand

plugin_names = {}


plugin_migrations = {}
data_migration = {}
plugin_prefixes = []

# Bootstrap 4
if "djangocms_bootstrap4" in apps.all_models:
    from djangocms_frontend.management import bootstrap4_migration

    plugin_migrations.update(bootstrap4_migration.plugin_migrations)
    data_migration.update(bootstrap4_migration.data_migration)
    plugin_prefixes.append(bootstrap4_migration.plugin_prefix)
# Styled link
if "djangocms_styledlink" in apps.all_models:
    from djangocms_frontend.management import styled_link_migration

    plugin_migrations.update(styled_link_migration.plugin_migrations)
    data_migration.update(styled_link_migration.data_migration)
    plugin_prefixes.append(styled_link_migration.plugin_prefix)
# djangocms-icon
if "djangocms_icon" in apps.all_models:
    from djangocms_frontend.management import icon_migration

    plugin_migrations.update(icon_migration.plugin_migration)
    data_migration.update(icon_migration.data_migration)
    plugin_prefixes.append(icon_migration.plugin_prefix)

additional_migrations = getattr(
    settings, "DJANGOCMS_FRONTEND_ADDITIONAL_MIGRATIONS", None
)
if additional_migrations:
    if isinstance(additional_migrations, str):
        additional_migrations = [additional_migrations]

    for migration_module_path in additional_migrations:
        try:
            migration_module = importlib.import_module(migration_module_path)
            plugin_migrations.update(migration_module.plugin_migrations)
            data_migration.update(migration_module.data_migration)
            plugin_prefixes.append(migration_module.plugin_prefix)
        except ModuleNotFoundError:
            print(f"Warning: can not import migration module: {migration_module_path}.")


def migrate_to_djangocms_frontend(apps, schema_editor):
    cnt = 0
    for plugin_model, fields in plugin_migrations.items():
        old, new = plugin_model.split(" -> ")
        old_app, old_model = old.rsplit(".", 1)
        new_app, new_model = new.rsplit(".", 1)
        if old_app in apps.all_models:
            OldPluginModel = apps.get_model(old_app, old_model)
            NewPluginModel = apps.get_model(new_app, new_model)
            for obj in OldPluginModel.objects.all():
                #
                new_obj = NewPluginModel()
                new_obj_fields = [field.name for field in new_obj._meta.get_fields()]
                new_obj.id = obj.id
                new_obj.placeholder = obj.placeholder
                new_obj.parent = obj.parent
                new_obj.position = obj.position
                new_obj.language = obj.language
                new_obj.creation_date = obj.creation_date
                if hasattr(obj, "depth"):  # cms v3
                    new_obj.depth = obj.depth
                    new_obj.numchild = obj.numchild
                    new_obj.path = obj.path
                new_obj.plugin_type = (
                    plugin_names[new_model]
                    if new_model in plugin_names
                    else new_model + "Plugin"
                )
                # Add something like `new_obj.field_name = obj.field_name` for any field in the the new plugin
                for field in fields:
                    if field in data_migration:
                        data_migration[field](obj, new_obj)
                    else:
                        if " -> " in field:
                            old_field, new_field = field.split(" -> ")
                        else:
                            old_field, new_field = field, field
                        value = (
                            old_field[1:-1]
                            if old_field[0] == "("
                            else getattr(obj, old_field)
                        )
                        if value == "":
                            value = None
                        if new_field in new_obj_fields:
                            setattr(new_obj, new_field, value)
                        else:
                            if isinstance(value, models.Model):  # related field
                                value = {
                                    "model": "{}.{}".format(
                                        value._meta.app_label,
                                        value._meta.model_name,
                                    ),
                                    "pk": value.pk,
                                }
                            elif isinstance(
                                value, models.QuerySet
                            ):  # related many field
                                value = {
                                    "model": "{}.{}".format(
                                        value.model._meta.app_label,
                                        value.model._meta.model_name,
                                    ),
                                    "p_keys": list(value.values_list("pk", flat=True)),
                                }
                            new_obj.config[new_field] = value
                new_obj.save()
                # Now delete old plugin from its table w/o checking for child plugins
                with connection.cursor() as cursor:
                    sql_command = f"DELETE FROM `{obj._meta.db_table}` WHERE cmsplugin_ptr_id={obj.id};"
                    if connection.vendor not in ("mysql", "sqlite"):
                        # ANSI: Use double quotes instead of backticks
                        sql_command = sql_command.replace("`", '"')
                    cursor.execute(sql_command)
                cnt += 1
                print(f"{cnt:7}", end="\r")
                # Copy any many to many field after save:`new_plugin.many2many.set(old_plugin.many2many.all())`
        else:
            print(f"{old_app} not installed.")
    if cnt:
        print(f"Migrated {cnt} plugins.")
    return cnt


blog_example = """
You have djangocms_blog installed. Consider adding the following
lines to your settings.py:

    from django.utils.translation import getext_lazy as _

    DJANGOCMS_FRONTEND_LINK_MODELS = [
        {
            "type": _("Blog pages"),
            "class_path": "djangocms_blog.models.Post",
            "filter": {"publish": True, "app_config_id": 1},
            "search": "translations__title",
        },
    ]

This will allow editors to directly link to blog posts. Also,
blog posts can easily link amongst themselves.

See for https://djangocms-frontend.readthedocs.io/en/latest/howto_guides.html
for more information.
"""

doc_reference = """
You may want to consider allowing editors to directly link to
a page generated by these models.

See for https://djangocms-frontend.readthedocs.io/en/latest/howto_guides.html
for more information.
"""


class Migrate(SubcommandsCommand):
    help = "Migrates plugins djangocms_bootstrap4 to djangocms_frontend"
    command_name = "migrate"

    def handle(self, *args, **options):
        tables = connection.introspection.table_names()
        if "djangocms_frontend_frontenduiitem" not in tables:
            self.stdout.write(
                self.style.ERROR(
                    "I cannot find djangocms-frontend's tables in the database. Did you run\n"
                    "./manage.py migrate ?"
                )
            )
            return

        self.migrate_to_djangocms_frontend()
        if getattr(settings, "DJANGOCMS_FRONTEND_LINK_MODELS", None) is None:
            self.check_for_link_targets()

    def migrate_to_djangocms_frontend(self):
        from cms.models.pluginmodel import CMSPlugin

        self.stdout.write(self.style.SUCCESS("Migrating plugins"))
        self.stdout.write(self.style.SUCCESS("================="))
        changes = migrate_to_djangocms_frontend(apps, None)
        not_migrated = []
        for plugin in CMSPlugin.objects.all():
            if any([prefix in plugin.plugin_type for prefix in plugin_prefixes]):
                if plugin.plugin_type not in not_migrated:
                    not_migrated.append(plugin.plugin_type)
                    self.stdout.write(
                        self.style.WARNING(f"{plugin.plugin_type} not migrated.")
                    )
        if not not_migrated:
            if changes:
                self.stdout.write(self.style.SUCCESS("Successfully migrated plugins."))
            else:
                self.stdout.write("Nothing to migrate")

    def check_for_link_targets(self):
        self.stdout.write()
        self.stdout.write(
            self.style.SUCCESS(
                "Checking installed apps for potential link destinations"
            )
        )
        self.stdout.write(
            self.style.SUCCESS(
                "======================================================="
            )
        )
        blog = False
        count = 0
        for app, app_models in apps.all_models.items():
            if app != "cms":
                for _, model in app_models.items():
                    if hasattr(model, "get_absolute_url"):
                        count += 1
                        self.stdout.write(
                            self.style.NOTICE(
                                f"App {app}'s {model.__name__} model is a suitable link destination."
                            )
                        )
            if app == "djangocms_blog":
                blog = True
        if count:
            self.stdout.write(
                self.style.SUCCESS(f"{count} potential link destinations found.")
            )
            if blog:
                self.stdout.write(self.style.WARNING(blog_example))
            else:
                self.stdout.write(self.style.WARNING(doc_reference))

        else:
            self.stdout.write("No further link destinations found. Setup complete.")



from django.utils.translation import gettext_lazy as _

from djangocms_frontend.models import FrontendUIItem


class Alert(FrontendUIItem):
    """
    Components > "Alerts" Plugin
    https://getbootstrap.com/docs/5.0/components/alerts/
    """

    class Meta:
        proxy = True
        verbose_name = _("Alert")

    def get_short_description(self):
        return f"({self.alert_context})"



from cms.plugin_pool import plugin_pool
from django.utils.translation import gettext_lazy as _

from ... import settings
from ...cms_plugins import CMSUIPlugin
from ...common.attributes import AttributesMixin
from ...common.responsive import ResponsiveMixin
from ...common.spacing import MarginMixin, PaddingMixin
from .. import listgroup
from . import forms, models

mixin_factory = settings.get_renderer(listgroup)


@plugin_pool.register_plugin
class ListGroupPlugin(
    mixin_factory("ListGroup"),
    AttributesMixin,
    ResponsiveMixin,
    MarginMixin,
    CMSUIPlugin,
):
    """
    Components > "List Group" Plugin
    https://getbootstrap.com/docs/5.0/components/list-group/
    """

    name = _("List group")
    module = _("Frontend")
    model = models.ListGroup
    form = forms.ListGroupForm
    change_form_template = "djangocms_frontend/admin/list-group.html"
    allow_children = True
    child_classes = ["ListGroupItemPlugin", "LinkPlugin"]
    # TODO consider linking to tab-content

    fieldsets = [
        (None, {"fields": ("list_group_flush",)}),
    ]


@plugin_pool.register_plugin
class ListGroupItemPlugin(
    mixin_factory("ListGroupItem"),
    AttributesMixin,
    ResponsiveMixin,
    PaddingMixin,
    CMSUIPlugin,
):
    """
    Components > "List Group Item" Plugin
    https://getbootstrap.com/docs/5.0/components/list-group/
    """

    name = _("List item")
    module = _("Frontend")
    model = models.ListGroupItem
    form = forms.ListGroupItemForm
    change_form_template = "djangocms_frontend/admin/list-group.html"
    allow_children = True
    parent_classes = ["ListGroupPlugin"]

    fieldsets = [
        (
            None,
            {
                "fields": (
                    "simple_content",
                    "list_context",
                    "list_state",
                )
            },
        ),
    ]



from django.test import TestCase

from djangocms_frontend.contrib.collapse.models import (
    Collapse,
    CollapseContainer,
    CollapseTrigger,
)


class B5CollapseModelTestCase(TestCase):
    def test_collapse_instance(self):
        instance = Collapse.objects.create()
        self.assertEqual(str(instance), "Collapse (1)")
        self.assertEqual(instance.get_short_description(), "(collapse-1)")

    def test_collapse_trigger_instance(self):
        instance = CollapseTrigger.objects.create(
            config={"trigger_identifier": "Monty Python"}
        )
        self.assertEqual(str(instance), "CollapseTrigger (1)")
        self.assertEqual(instance.get_short_description(), "(Monty Python)")

    def test_collapse_container_instance(self):
        instance = CollapseContainer.objects.create(
            config={"container_identifier": "Monty Python"}
        )
        self.assertEqual(str(instance), "CollapseContainer (1)")
        self.assertEqual(instance.get_short_description(), "(Monty Python)")



from django import forms
from django.utils.translation import gettext_lazy as _
from entangled.forms import EntangledModelForm

from djangocms_frontend.fields import (
    AttributesFormField,
    ColoredButtonGroup,
    TagTypeFormField,
)

from ... import settings
from ...common.background import BackgroundFormMixin
from ...common.responsive import ResponsiveFormMixin
from ...common.spacing import SpacingFormMixin
from ...helpers import first_choice
from ...models import FrontendUIItem
from ...settings import COLOR_STYLE_CHOICES
from .conf import ICON_SIZE_CHOICES, ICON_TAG_TYPES
from .fields import IconPickerField


class IconForm(BackgroundFormMixin, ResponsiveFormMixin, SpacingFormMixin, EntangledModelForm):
    """
    Layout > "Media" Plugin
    http://getbootstrap.com/docs/4.0/layout/media-object/
    """

    class Meta:
        model = FrontendUIItem
        entangled_fields = {
            "config": [
                "icon",
                "icon_size",
                "icon_foreground",
                "icon_rounded",
                "attributes",
            ]
        }
        untangled_fields = (
            "tag_type",
        )

    icon = IconPickerField()
    icon_size = forms.ChoiceField(
        label=_("Icon size"),
        choices=ICON_SIZE_CHOICES,
        initial=first_choice(ICON_SIZE_CHOICES),
        required=False,
    )
    icon_foreground = forms.ChoiceField(
        label=_("Foreground context"),
        choices=settings.EMPTY_CHOICE + COLOR_STYLE_CHOICES,
        initial=settings.EMPTY_CHOICE[0][0],
        widget=ColoredButtonGroup(),
        required=False,
    )
    icon_rounded = forms.BooleanField(
        label=_("Circular icon"),
        required=False,
    )
    attributes = AttributesFormField()
    tag_type = TagTypeFormField(
        choices=ICON_TAG_TYPES,
        initial=first_choice(ICON_TAG_TYPES),
    )



from django.test import TestCase

from djangocms_frontend.contrib.alert.forms import AlertForm
from djangocms_frontend.contrib.alert.models import Alert


class AlertModelTestCase(TestCase):
    def test_instance(self):
        instance = Alert.objects.create()
        instance.initialize_from_form(AlertForm).save()
        self.assertEqual(str(instance), "Alert (1)")
        self.assertEqual(instance.get_short_description(), "(primary)")



from django.test import TestCase

from djangocms_frontend.settings import (
    ALIGN_CHOICES,
    COLOR_STYLE_CHOICES,
    DEVICE_CHOICES,
    DEVICE_SIZES,
    TAG_CHOICES,
)


class B5ConstantsTestCase(TestCase):
    # make sure to update the documentation
    # when changing any values in these

    def test_constants(self):
        self.assertEqual(
            DEVICE_CHOICES,
            (
                ("xs", "Extra small"),  # default <576px
                ("sm", "Small"),  # default ≥576px
                ("md", "Medium"),  # default ≥768px
                ("lg", "Large"),  # default ≥992px
                ("xl", "Extra large"),  # default ≥1200px
                ("xxl", "Extra-extra large"),
            ),
        )
        self.assertEqual(DEVICE_SIZES, ("xs", "sm", "md", "lg", "xl", "xxl"))
        self.assertEqual(
            TAG_CHOICES,
            (
                ("div", "div"),
                ("section", "section"),
                ("article", "article"),
                ("header", "header"),
                ("footer", "footer"),
                ("aside", "aside"),
            ),
        )
        self.assertEqual(
            COLOR_STYLE_CHOICES,
            (
                ("primary", "Primary"),
                ("secondary", "Secondary"),
                ("success", "Success"),
                ("danger", "Danger"),
                ("warning", "Warning"),
                ("info", "Info"),
                ("light", "Light"),
                ("dark", "Dark"),
            ),
        )
        self.assertEqual(
            ALIGN_CHOICES,
            (
                ("start", "Left"),
                ("center", "Center"),
                ("end", "Right"),
            ),
        )



from django.utils.translation import gettext as _

from djangocms_frontend.models import FrontendUIItem


class Icon(FrontendUIItem):
    """
    Adds icons from configurable icon picker
    https://github.com/migliori/universal-icon-picker
    """

    class Meta:
        proxy = True
        verbose_name = _("Icon")

    def get_short_description(self):
        return self.config.get("icon", {}).get("iconClass", _("undefined"))



import os

from django.utils.html import strip_tags
from django.utils.translation import gettext_lazy as _

from djangocms_frontend.models import FrontendUIItem

from ..image.models import ImageMixin
from ..link.models import GetLinkMixin


class Carousel(FrontendUIItem):
    """
    Components > "Carousel" Plugin
    https://getbootstrap.com/docs/5.0/components/carousel/
    """

    class Meta:
        proxy = True
        verbose_name = _("Carousel")

    def get_short_description(self):
        text = f"({self.template})"
        text += " {}: {}".format(_("Interval"), self.carousel_interval)
        text += ", {}: {}".format(_("Controls"), self.carousel_controls)
        text += ", {}: {}".format(_("Indicators"), self.carousel_indicators)
        text += ", {}: {}".format(_("Keyboard"), self.carousel_keyboard)
        text += ", {}: {}".format(_("Pause"), self.carousel_pause)
        text += ", {}: {}".format(_("Ride"), self.carousel_ride)
        text += ", {}: {}".format(_("Wrap"), self.carousel_wrap)
        return text


class CarouselSlide(GetLinkMixin, ImageMixin, FrontendUIItem):
    """
    Components > "Slide" Plugin
    https://getbootstrap.com/docs/5.0/components/carousel/
    """

    class Meta:
        proxy = True
        verbose_name = _("Carousel slide")

    image_field = "carousel_image"

    def get_short_description(self):
        image_text = content_text = ""

        if self.carousel_image:
            if self.rel_image is None:
                image_text = _("<file is missing>")
            elif self.rel_image.name:
                image_text = self.rel_image.name
            elif (
                self.rel_image.original_filename
                and os.path.split(self.rel_image.original_filename)[1]
            ):
                image_text = os.path.split(self.rel_image.original_filename)[1]
            else:
                image_text = "Image"
        if self.carousel_content:
            text = strip_tags(self.carousel_content).strip()
            if len(text) > 100:
                content_text = f"{text[:100]}..."
            else:
                content_text = f"{text}"

        if image_text and content_text:
            return f"{image_text} ({content_text})"
        else:
            return image_text or content_text



from cms.plugin_pool import plugin_pool
from django.utils.translation import gettext_lazy as _

from ... import settings
from ...cms_plugins import CMSUIPlugin
from ...common.attributes import AttributesMixin
from ...common.responsive import ResponsiveMixin
from ...common.spacing import MarginMixin
from .. import image
from ..link.cms_plugins import LinkPluginMixin
from . import forms, models

mixin_factory = settings.get_renderer(image)


@plugin_pool.register_plugin
class ImagePlugin(
    mixin_factory("Image"),
    AttributesMixin,
    ResponsiveMixin,
    MarginMixin,
    LinkPluginMixin,
    CMSUIPlugin,
):
    """
    Content > "Image" Plugin
    https://getbootstrap.com/docs/5.0/content/images/
    """

    name = _("Picture / Image")
    module = _("Frontend")

    model = models.Image
    form = forms.ImageForm

    change_form_template = "djangocms_frontend/admin/image.html"

    fieldsets = [
        (
            None,
            {
                "fields": (
                    "template",
                    "picture",
                    "external_picture",
                    (
                        "picture_fluid",
                        "lazy_loading",
                        "picture_rounded",
                        "picture_thumbnail",
                    ),
                )
            },
        ),
        (
            _("Format"),
            {
                "classes": ("collapse",),
                "fields": (
                    "use_responsive_image",
                    ("width", "height"),
                    "alignment",
                ),
            },
        ),
        (
            _("Cropping"),
            {
                "classes": ("collapse",),
                "fields": (
                    ("use_automatic_scaling", "use_no_cropping"),
                    ("use_crop", "use_upscale"),
                    "thumbnail_options",
                ),
            },
        ),
    ]
    link_fieldset_position = -1

    def get_render_template(self, context, instance, placeholder):
        return f"djangocms_frontend/{settings.framework}/{instance.template}/image.html"

    def render(self, context, instance, placeholder):
        if instance.config.get("lazy_loading", False):
            instance.add_attribute("loading", "lazy")
        return super().render(context, instance, placeholder)



from cms.api import add_plugin
from cms.test_utils.testcases import CMSTestCase

from djangocms_frontend.contrib.utilities.cms_plugins import (
    HeadingPlugin,
    SpacingPlugin,
    TOCPlugin,
)
from djangocms_frontend.contrib.utilities.forms import TableOfContentsForm

from ..fixtures import TestFixture


class UtilitiesPluginTestCase(TestFixture, CMSTestCase):
    def test_spacing(self):
        add_plugin(
            placeholder=self.placeholder,
            plugin_type=SpacingPlugin.__name__,
            language=self.language,
            config={
                "space_property": "m",
                "space_size": 0,
                "space_device": "xs",
                "space_sides": "",
            },
        )
        self.publish(self.page, self.language)

        with self.login_user_context(self.superuser):
            response = self.client.get(self.request_url)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, '<div class="m-0">')

    def test_heading(self):
        add_plugin(
            placeholder=self.placeholder,
            plugin_type=TOCPlugin.__name__,
            language=self.language,
            config=dict(list_attributes={"class": "empty-toc"}),
        ).initialize_from_form(TableOfContentsForm).save()
        add_plugin(
            placeholder=self.placeholder,
            plugin_type=HeadingPlugin.__name__,
            language=self.language,
            config={
                "heading_level": "h2",
                "heading": "Welcome to django CMS!",
                "heading_id": "id1",
            },
        )
        add_plugin(
            placeholder=self.placeholder,
            plugin_type=HeadingPlugin.__name__,
            language=self.language,
            config={
                "heading_level": "h3",
                "heading": "How you can benefit from django CMS",
                "heading_id": "id2",
            },
        )
        add_plugin(
            placeholder=self.placeholder,
            plugin_type=HeadingPlugin.__name__,
            language=self.language,
            config={
                "heading_level": "h2",
                "heading": "Great, isn't it?",
                "heading_id": "id3",
            },
        )
        add_plugin(
            placeholder=self.placeholder,
            plugin_type=TOCPlugin.__name__,
            language=self.language,
            config=dict(list_attributes={"class": "test-class"}),
        ).initialize_from_form(TableOfContentsForm).save()
        self.publish(self.page, self.language)

        with self.login_user_context(self.superuser):
            response = self.client.get(self.request_url)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, '<ul class="empty-toc"></ul>')
        self.assertContains(response, '<h2 id="id1">Welcome to django CMS!</h2>')
        self.assertContains(
            response, '<h3 id="id2">How you can benefit from django CMS</h3>'
        )
        self.assertContains(response, '<ul class="test-class"><li ><a href="#id2" >')
        self.assertContains(response, '<a href="#id1" >')
        self.assertContains(response, '<a href="#id2" >')



from cms.api import add_plugin
from cms.test_utils.testcases import CMSTestCase

from djangocms_frontend.contrib.navigation.cms_plugins import (
    NavContainerPlugin,
    NavigationPlugin,
    NavLinkPlugin,
    PageTreePlugin,
)
from djangocms_frontend.contrib.navigation.forms import (
    NavContainerForm,
    NavigationForm,
    NavLinkForm,
    PageTreeForm,
)

from ..fixtures import TestFixture


class NavigationPluginTestCase(TestFixture, CMSTestCase):
    def test_plugin(self):
        nav = add_plugin(
            placeholder=self.placeholder,
            plugin_type=NavigationPlugin.__name__,
            language=self.language,
            config=dict(
                navbar_breakpoint="xl",
                navbar_container=True,
            ),
        ).initialize_from_form(NavigationForm)
        nav.save()
        self.publish(self.page, self.language)

        with self.login_user_context(self.superuser):
            response = self.client.get(self.request_url)
        self.assertEqual(response.status_code, 200)
        self.assertTrue(
            '<nav class="navbar-expand-xl navbar navbar-light">'
            in response.content.decode("utf-8")
            or '<nav class="navbar-expand-xl navbar-light navbar">'
            in response.content.decode("utf-8")
            or '<nav class="navbar navbar-light navbar-expand-xl">'
            in response.content.decode("utf-8")
            or '<nav class="navbar navbar-expand-xl navbar-light">'
            in response.content.decode("utf-8")
            or '<nav class="navbar-light navbar navbar-expand-xl">'
            in response.content.decode("utf-8")
            or '<nav class="navbar-light navbar-expand-xl navbar">'
            in response.content.decode("utf-8")
        )
        self.assertContains(response, '<div class="container"><ul class="navbar-nav"></ul></div>')

        # add more options
        container = add_plugin(
            placeholder=self.placeholder,
            plugin_type=NavContainerPlugin.__name__,
            target=nav,
            language=self.language,
            config=dict(),
        )
        container.initialize_from_form(NavContainerForm).save()
        self.publish(self.page, self.language)

        with self.login_user_context(self.superuser):
            response = self.client.get(self.request_url)

        self.assertEqual(response.status_code, 200)
        self.assertContains(response, '<ul class="navbar-nav"></ul>')
        self.assertContains(response, '<button class="navbar-toggler"')
        self.assertContains(response, 'id="nav')

        plugin = add_plugin(
            placeholder=self.placeholder,
            plugin_type=NavLinkPlugin.__name__,
            language=self.language,
            target=container,
            config=dict(
                internal_link=dict(model="cms.page", pk=self.page.id),
                link_context="primary",
                link_type="btn",
                name="django CMS rocks!",
            ),
        )
        plugin.initialize_from_form(NavLinkForm).save()
        self.publish(self.page, self.language)

        with self.login_user_context(self.superuser):
            response = self.client.get(self.request_url)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "btn-primary")
        self.assertContains(response, "nav-link")
        self.assertContains(response, 'href="/content/"')

        plugin = add_plugin(
            placeholder=self.placeholder,
            plugin_type=PageTreePlugin.__name__,
            language=self.language,
            target=container,
            config=dict(),
        )
        plugin.initialize_from_form(PageTreeForm).save()
        self.publish(self.page, self.language)

        with self.login_user_context(self.superuser):
            response = self.client.get(self.request_url)
        self.assertEqual(response.status_code, 200)



from django.test import TestCase

from djangocms_frontend.contrib.badge.forms import BadgeForm
from djangocms_frontend.contrib.badge.models import Badge


class BadgeModelTestCase(TestCase):
    def test_instance(self):
        instance = Badge.objects.create().initialize_from_form(BadgeForm)
        self.assertEqual(str(instance), "Badge (1)")
        self.assertEqual(instance.get_short_description(), "(primary)")



from django.test import TestCase

from djangocms_frontend.contrib.navigation.forms import (
    NavigationForm,
    NavLinkForm,
    PageTreeForm,
)
from djangocms_frontend.contrib.navigation.models import Navigation, NavLink, PageTree


class NavigationTestCase(TestCase):
    def test_instance(self):
        instance = Navigation.objects.create().initialize_from_form(NavigationForm)
        self.assertEqual(str(instance), "Navigation (1)")
        self.assertEqual(instance.get_short_description(), "(light)")


class PageTreeTestCase(TestCase):
    def test_instance(self):
        instance = PageTree.objects.create().initialize_from_form(PageTreeForm)
        self.assertEqual(str(instance), "PageTree (1)")
        self.assertEqual(instance.get_short_description(), "")


class NavLinkTestCase(TestCase):
    def test_instance(self):
        instance = NavLink.objects.create().initialize_from_form(NavLinkForm)
        self.assertEqual(str(instance), "NavLink (1)")
        self.assertEqual(instance.get_short_description(), "<link is missing>")



class AccordionRenderMixin:
    render_template = "djangocms_frontend/bootstrap5/accordion.html"

    def render(self, context, instance, placeholder):
        instance.add_classes("accordion")
        if instance.accordion_flush:
            instance.add_classes("accordion-flush")
        return super().render(context, instance, placeholder)


class AccordionItemRenderMixin:
    render_template = "djangocms_frontend/bootstrap5/accordion_item.html"

    def render(self, context, instance, placeholder):
        instance.add_classes("accordion-collapse collapse")
        if instance.accordion_item_open:
            instance.add_classes("show")
        instance.font_size = (
            context["parent"]
            .config.get("accordion_header_type", "")
            .replace("h", "fs-")
        )
        return super().render(context, instance, placeholder)



from cms.plugin_pool import plugin_pool
from django.utils.translation import gettext_lazy as _

from ... import settings
from ...cms_plugins import CMSUIPlugin
from ...common.attributes import AttributesMixin
from ...common.background import BackgroundMixin
from ...common.responsive import ResponsiveMixin
from ...common.spacing import SpacingMixin
from .. import content
from . import forms, models

mixin_factory = settings.get_renderer(content)


@plugin_pool.register_plugin
class CodePlugin(
    mixin_factory("Code"),
    AttributesMixin,
    ResponsiveMixin,
    SpacingMixin,
    BackgroundMixin,
    CMSUIPlugin,
):
    """
    Content > "Code" Plugin
    https://getbootstrap.com/docs/5.0/content/code/
    """

    name = _("Code")
    module = _("Frontend")
    model = models.CodeBlock
    form = forms.CodeForm
    change_form_template = "djangocms_frontend/admin/code.html"

    fieldsets = [
        (
            None,
            {
                "fields": (
                    "code_content",
                    "code_type",
                )
            },
        ),
    ]


@plugin_pool.register_plugin
class BlockquotePlugin(
    mixin_factory("Blockquote"),
    AttributesMixin,
    ResponsiveMixin,
    SpacingMixin,
    BackgroundMixin,
    CMSUIPlugin,
):
    """
    Content > "Blockquote" Plugin
    https://getbootstrap.com/docs/5.0/content/typography/#blockquotes
    """

    name = _("Blockquote")
    module = _("Frontend")
    model = models.Blockquote
    form = forms.BlockquoteForm
    change_form_template = "djangocms_frontend/admin/blockquote.html"
    allow_children = True

    fieldsets = [
        (
            None,
            {
                "fields": (
                    "quote_content",
                    "quote_origin",
                    "quote_alignment",
                )
            },
        ),
    ]


@plugin_pool.register_plugin
class FigurePlugin(
    mixin_factory("Figure"),
    AttributesMixin,
    ResponsiveMixin,
    SpacingMixin,
    BackgroundMixin,
    CMSUIPlugin,
):
    """
    Content > "Figure" Plugin
    https://getbootstrap.com/docs/5.0/content/figures/
    """

    name = _("Figure")
    module = _("Frontend")
    model = models.Figure
    form = forms.FigureForm
    change_form_template = "djangocms_frontend/admin/figure.html"
    allow_children = True

    fieldsets = [
        (
            None,
            {
                "fields": (
                    "figure_caption",
                    "figure_alignment",
                )
            },
        ),
    ]



from django import forms
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _
from entangled.forms import EntangledModelFormMixin


class TitleWidget(forms.MultiWidget):  # lgtm [py/missing-call-to-init]
    def __init__(self, *args, **kwargs):
        kwargs.setdefault(
            "widgets",
            (
                forms.CheckboxInput(),
                forms.TextInput(),
            ),
        )
        super().__init__(*args, **kwargs)

    def decompress(self, value):
        if isinstance(value, dict):
            return [value.get("show", False), value.get("title", "")]
        return [False, ""]


class TitleField(forms.MultiValueField):  # lgtm [py/missing-call-to-init]
    def __init__(self, *args, **kwargs):
        kwargs.setdefault(
            "fields",
            (
                forms.BooleanField(required=False),
                forms.CharField(required=False),
            ),
        )
        kwargs.setdefault("require_all_fields", False)
        kwargs.setdefault("widget", TitleWidget())
        super().__init__(*args, **kwargs)

    def clean(self, value):
        if value[0] and not value[1]:
            raise ValidationError(
                _("Please add a title if you want to publish it."), code="incomplete"
            )
        return super().clean(value)

    def compress(self, data_list):
        return dict(show=data_list[0], title=data_list[1])


class TitleMixin:
    def render(self, context, instance, placeholder):
        if instance.config.get("plugin_title", {}).get("show", False):
            instance.add_attribute("title", instance.plugin_title.get("title", ""))
        return super().render(context, instance, placeholder)


class TitleFormMixin(EntangledModelFormMixin):
    class Meta:
        entangled_fields = {
            "config": [
                "plugin_title",
            ]
        }

    plugin_title = TitleField(
        label=_("Title"),
        required=False,
        help_text=_(
            "Optional title of the plugin for easier identification. "
            "Its <code>title</code> attribute "
            "will only be set if the checkbox is selected."
        ),
    )



class ListGroupRenderMixin:
    def render(self, context, instance, placeholder):
        instance.add_classes("list-group")
        if instance.list_group_flush:
            instance.add_classes("list-group-flush")
        return super().render(context, instance, placeholder)


class ListGroupItemRenderMixin:
    def render(self, context, instance, placeholder):
        instance.add_classes("list-group-item")
        if instance.list_context:
            instance.add_classes(f"list-group-item-{instance.list_context}")
        if instance.list_state:
            instance.add_classes(instance.list_state)

        return super().render(context, instance, placeholder)



from djangocms_frontend.contrib.grid.frameworks.bootstrap5 import (
    get_row_cols_grid_values,
)


class CardRenderMixin:
    render_template = "djangocms_frontend/bootstrap5/card.html"

    def render(self, context, instance, placeholder):
        instance.add_classes("card")
        if instance.config.get("card_outline", None):
            instance.add_classes(f"border-{instance.card_outline}")
        if instance.card_alignment:
            instance.add_classes(f"text-{instance.card_alignment}")
        if instance.config.get("card_text_color", None):
            instance.add_classes(f"text-{instance.card_text_color}")
        if instance.config.get("card_full_height", None):
            instance.add_classes("h-100")
        if instance.parent and instance.parent.plugin_type == "CardLayoutPlugin":
            if instance.parent.get_plugin_instance()[0].card_type == "row":
                instance.add_classes("h-100")
        return super().render(context, instance, placeholder)


class CardInnerRenderMixin:
    def render(self, context, instance, placeholder):
        instance.add_classes(instance.inner_type)
        if getattr(instance, "text_alignment", None):
            instance.add_classes(f"text-{instance.text_alignment}")
        return super().render(context, instance, placeholder)


class CardLayoutRenderMixin:
    def render(self, context, instance, placeholder):
        instance.add_classes(instance.card_type)
        instance.add_classes(get_row_cols_grid_values(instance))
        return super().render(context, instance, placeholder)
